"""
=============================================================================

    This module contains tests for the functionalities in the tuner module.

=============================================================================

    Copyright 2022 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

import unittest
from typing import Sequence

import numpy as np
import pennylane as qml
from numpy.random import RandomState
from numpy.typing import NDArray
from sklearn.datasets import (
    load_digits,
    make_classification,
    make_moons,
    make_regression,
)

from aqmlator.qml import QNNBinaryClassifier
from aqmlator.tuner import HyperparameterTuner, MLTaskType, ModelFinder


class TestModelFinder(unittest.TestCase):
    """
    This is a `TestCase` for the `ModelFinder` class.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.

        :Note:
            There's no return value and only a single method to test, so we strive
            to be as minimalistic as possible.
        """

        x: Sequence[Sequence[float]]
        y: Sequence[int]

        x, y = make_moons(
            n_samples=10,
            shuffle=True,
            noise=0.1,
            random_state=RandomState(0),
        )

        reg_x, reg_y = make_regression(  # pylint: disable=unbalanced-tuple-unpacking
            n_samples=10,
            n_features=2,
            n_informative=2,
            n_targets=1,
            random_state=RandomState(0),
        )

        cls_x, cls_y = make_classification(
            n_samples=10,
            n_features=2,
            n_informative=2,
            n_redundant=0,
            n_repeated=0,
            n_classes=3,
            n_clusters_per_class=1,
            random_state=RandomState(0),
        )

        n_qubits: int = 2
        n_seeds: int = 2
        n_trials: int = 4
        n_epochs: int = 3

        dev: qml.Device = qml.device("lightning.qubit", wires=n_qubits)

        self.binary_classifier_finder: ModelFinder = ModelFinder(
            task_type=MLTaskType.BINARY_CLASSIFICATION,
            features=x,
            classes=y,
            n_cores=1,
            n_trials=n_trials,
            n_seeds=n_seeds,
            n_epochs=n_epochs,
            device=dev,
        )

        self.classifier_finder: ModelFinder = ModelFinder(
            task_type=MLTaskType.CLASSIFICATION,
            features=cls_x,
            classes=cls_y,
            n_cores=1,
            n_trials=n_trials,
            n_seeds=n_seeds,
            n_epochs=n_epochs,
            device=dev,
        )

        self.linear_regressor_finder: ModelFinder = ModelFinder(
            task_type=MLTaskType.REGRESSION,
            features=reg_x,
            classes=reg_y,
            n_cores=1,
            n_trials=n_trials,
            n_seeds=n_seeds,
            n_epochs=n_epochs,
            device=dev,
        )

        n_classes: int = 2
        grp_x: NDArray[np.float32]
        grp_x, _ = load_digits(n_class=n_classes, return_X_y=True)
        grp_x = grp_x.reshape(len(grp_x), 1, 8, 8)
        grp_x /= 16

        grp_x = grp_x.astype(np.float32)

        self.grouping_model_finder: ModelFinder = ModelFinder(
            task_type=MLTaskType.GROUPING,
            features=grp_x,
            n_cores=1,
            n_trials=n_trials,
            n_seeds=n_seeds,
            n_epochs=n_epochs,
        )

    def test_binary_classification_model_finding(self) -> None:
        """
        Tests if `ModelFinder` finds a binary classification model.
        """
        self.binary_classifier_finder.find_model()

    def test_classification_model_finding(self) -> None:
        """
        Tests if `ModelFinder` finds a linear regression model.
        """
        self.classifier_finder.find_model()

    def test_linear_regression_model_finding(self) -> None:
        """
        Tests if `ModelFinder` finds a linear regression model.
        """
        self.linear_regressor_finder.find_model()

    def test_grouping_model_finding(self) -> None:
        """
        Tests if `ModelFinder` finds a grouping model.
        """
        self.grouping_model_finder.find_model()


class TestHyperparameterTuner(unittest.TestCase):
    """
    This is a `TestCase` for the `HyperparameterTuner` class.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.

        :Note:
            There's no return value and only a single method to test, so we strive
            to be as minimalistic as possible.
        """
        x: Sequence[Sequence[float]]
        y: Sequence[int]

        x, y = make_moons(
            n_samples=100,
            shuffle=True,
            noise=0.1,
            random_state=RandomState(0),
        )

        n_seeds: int = 2
        n_trials: int = 2
        n_qubits: int = 2

        dev: qml.Device = qml.device("lightning.qubit", wires=n_qubits)

        classifier: QNNBinaryClassifier = QNNBinaryClassifier(2, 20, 5, device=dev)

        self.tuner: HyperparameterTuner = HyperparameterTuner(
            x,
            y,
            classifier,
            n_seeds=n_seeds,
            n_trials=n_trials,
        )

    def test_hyperparameter_tuner_running(self) -> None:
        """
        Tests if `HyperparameterTuner` runs.
        """
        self.tuner.find_hyperparameters()



"""
=============================================================================

    This module contains the functionalities required for receiving the data from
    the user and parsing them into a format that is used throughout the rest of the
    library.

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"


import csv
import locale
from abc import ABC, abstractmethod
from dataclasses import dataclass
from os.path import exists
from typing import Any, List, Tuple, Union


@dataclass(init=True, repr=True)
class LearningDatum:
    """
    A general class for holding the user-passed learning data.
    """

    datum_attributes: Tuple[Union[float, str], ...]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, LearningDatum):
            return tuple(self.datum_attributes) == tuple(other.datum_attributes)
        return NotImplemented


@dataclass(init=True, repr=True)
class SupervisedLearningDatum(LearningDatum):
    """
    A class for holding user-passed data for supervised learning. It holds additional
    information about the class / value of the target function.
    """

    datum_target: Union[float, str, int]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SupervisedLearningDatum):
            return (
                tuple(self.datum_attributes) == tuple(other.datum_attributes)
                and self.datum_target == other.datum_target
            )
        return NotImplemented


class DataReceiverInterface(ABC):
    """
    An interface class for all the data receivers.
    """

    @abstractmethod
    def receive_data(self, data_file_path: str) -> List[LearningDatum]:
        """
        The main method of the data receivers. It handles the initial data

        :param data_file_path:
            The path to the file containing the data.
        """
        raise NotImplementedError


class CSVDataReceiver(DataReceiverInterface):
    """
    This class is meant to receive the data given in a CSV format.
    """

    def __init__(
        self, data_separator: str = ",", target_index: Union[None, int] = None
    ) -> None:
        """
        A constructor for the CSVDataReceiver.

        :param data_separator:
            A symbol in the file used to separate the data.
        :param target_index:
            Index of the target value for supervised learning tasks. If set to `None`,
            then the Receiver assumes unsupervised learning.

        """
        self._data_separator: str = data_separator
        self._target_index: Union[int, None] = target_index
        super().__init__()

    def receive_data(self, data_file_path: str) -> List[LearningDatum]:
        """
        The main method of the data receiver. It takes the path to the file containing
        the data and returns a list of LearningDatum objects representing the data in
        the CSV file.

        :param data_file_path:
            The path to the file containing the data.

        :raises FileNotFoundError:
            If the file does not exist.

        :return:
            A list of LearningDatum objects representing the data in the CSV file.
        """
        if not exists(data_file_path):
            raise FileNotFoundError

        data: List[LearningDatum] = []

        with open(
            data_file_path, "r", encoding=locale.getpreferredencoding()
        ) as csv_file:
            reader = csv.reader(
                csv_file,
                "excel",
                delimiter=self._data_separator,
                quoting=csv.QUOTE_NONNUMERIC,
            )

            for row in reader:
                row_data: List[Union[float, int, str]] = list(row)

                if self._target_index is None:
                    # Unsupervised learning
                    data.append(LearningDatum(tuple(row_data)))
                else:
                    # Supervised learning
                    target_value: Union[float, int, str] = row_data.pop(
                        self._target_index
                    )
                    data.append(SupervisedLearningDatum(tuple(row_data), target_value))

        return data



"""
=============================================================================

    This module contains methods for handling database operations related to
    aqmlator.

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

import locale
import os
import sqlite3
from subprocess import Popen

__author__ = "Tomasz Rybotycki"


def _dump_postgres_base(dump_file_name: str = "aqmlatorDump.sql") -> None:
    """
    Dumps postgres database to a file with given name.

    :param dump_file_name:
        Name of the file to dump the database into.
    """
    command: str = (
        "pg_dump --create --inserts -f "
        + dump_file_name
        + " -d "
        + os.environ["aqmlator_database_url"]
    )

    with Popen(command, shell=True) as proc:
        proc.wait()


def _parse_to_sqlite(sql_file: str = "aqmlatorDump.sql") -> None:
    """
    Parse the dumped postgres database into the sql format that can be used to
    initialize SQLite database.

    :param sql_file:
        A file containing dumped postgres database.
    """
    parsed_sql: str = ""

    with open(sql_file, "r", encoding=locale.getpreferredencoding()) as f:
        while True:
            line: str = f.readline()

            # Finish parsing when needed.
            if "PostgreSQL database dump complete" in line:
                break

            # Keep only CREATE TABLE and INSERT commands.
            if not ("CREATE TABLE" in line or "INSERT" in line):
                continue

            # Remove `public.` from the database names.
            line = line.replace("public.", "")

            parsed_sql += line

            while ";" not in line:
                line = f.readline()
                line = line.replace("public.", "")
                parsed_sql += line

    with open("parsed_" + sql_file, "w", encoding=locale.getpreferredencoding()) as f:
        f.write(parsed_sql)


def _initialize_sqlite_db(
    sql_file: str, sqlite_db_name: str = "aqmlatorSQLite.db"
) -> None:
    """
    Initializes the SQLite database from the given .sql file.

    :param sql_file:
        A file containing the `sql` script to initialize the SQLite database.
    :param sqlite_db_name:
        The name of the resultant SQLite database file.
    """

    database = sqlite3.connect(sqlite_db_name)

    with open(sql_file, "r", encoding=locale.getpreferredencoding()) as f:
        database.executescript(f.read())

    database.close()


def export_data_to_sqlite_database(db_name: str = "aqmlatorSQLite.db") -> None:
    """
    Exports the aqmlator data to the SQLite database.

    :param db_name:
        Name of the resultant SQLite database.
    """
    sql_file_name: str = "temp.sql"
    _dump_postgres_base(sql_file_name)
    _parse_to_sqlite(sql_file_name)
    _initialize_sqlite_db("parsed_" + sql_file_name, db_name)

    # Remove sql files.
    os.remove(sql_file_name)
    os.remove("parsed_" + sql_file_name)



"""
=============================================================================

    This module contains the classes that use optuna for different kinds of
    optimizations - mainly model and hyperparameter.

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

import abc
import json
import uuid
from enum import StrEnum
from math import ceil, floor, prod, sqrt
from os import environ
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type, Union

import optuna
import pennylane as qml
import pennylane.numpy as np
import requests  # type: ignore[import-untyped]
from numpy.typing import NDArray
from optuna.samplers import TPESampler
from pennylane.optimize import (
    AdamOptimizer,
    GradientDescentOptimizer,
    NesterovMomentumOptimizer,
)
from pennylane.templates.embeddings import AmplitudeEmbedding, AngleEmbedding
from pennylane.templates.layers import BasicEntanglerLayers, StronglyEntanglingLayers
from sklearn.metrics import silhouette_score  # TR: It has bounds.
from torch import Tensor
from torch.utils.data import DataLoader

from aqmlator.qml import (
    QMLModel,
    QNNBinaryClassifier,
    QNNClassifier,
    QNNLinearRegression,
    QuantumKernelBinaryClassifier,
    RBMClustering,
)
from aqmlator.server import status_update_endpoint

# TODO TR:  Should those be global?

binary_classifiers: Dict[str, Dict[str, Any]] = {
    "QNN": {
        "constructor": QNNBinaryClassifier,
        "kwargs": {
            "batch_size": (15, 25),  # Might need to be data size-dependent instead.
        },
        "fixed_kwargs": {},
        "n_layers": (1, 3),
    },
    "QEK": {
        "constructor": QuantumKernelBinaryClassifier,
        "kwargs": {},
        "fixed_kwargs": {},
        "n_layers": (3, 5),
    },
}

regressors: Dict[str, Dict[str, Any]] = {
    "QNN": {
        "kwargs": {
            "batch_size": (15, 25),  # Might need to be data size-dependent instead.
        },
        "n_layers": (1, 3),
        "constructor": QNNLinearRegression,
    }
}

clustering: Dict[str, Dict[str, Any]] = {
    "RBM": {
        "constructor": RBMClustering,
        "kwargs": {
            "lbae_n_layers": (1, 2),  # TR: lbae_n_layers > 2 poses problems.
            "fireing_threshold": (0.6, 0.99),
        },
    }
}

data_embeddings: Dict[str, Dict[str, Any]] = {
    "ANGLE": {"constructor": AngleEmbedding, "kwargs": {}, "fixed_kwargs": {}},
    "AMPLITUDE": {
        "constructor": AmplitudeEmbedding,
        "kwargs": {},
        "fixed_kwargs": {"pad_with": 0, "normalize": True},
    },
}

optimizers: Dict[str, Dict[str, Any]] = {
    "NESTEROV": {
        "constructor": NesterovMomentumOptimizer,
        # Hyperparameters range taken from
        # https://cs231n.github.io/neural-networks-3/
        "kwargs": {
            "stepsize": {"min": 0.00001, "max": 0.1},
            "momentum": {"min": 0.5, "max": 0.9},
        },
    },
    "ADAM": {
        "constructor": AdamOptimizer,
        # Hyperparameters range taken from arXiv:1412.6980.
        "kwargs": {
            "stepsize": {
                "min": 0.00001,
                "max": 0.1,
            },
            "beta1": {
                "min": 0,
                "max": 0.9,
            },
            "beta2": {"min": 0.99, "max": 0.9999},
        },
    },
}

layer_types: Dict[str, Dict[str, Any]] = {
    "BASIC": {"constructor": BasicEntanglerLayers},
    "STRONGLY_ENTANGLING": {
        "constructor": StronglyEntanglingLayers,
    },
}


class MLTaskType(StrEnum):
    BINARY_CLASSIFICATION: str = "BINARY_CLASSIFICATION"
    CLASSIFICATION: str = "CLASSIFICATION"
    REGRESSION: str = "REGRESSION"
    GROUPING: str = "GROUPING"


class OptunaOptimizer(abc.ABC):
    """
    A class for all `optuna`-based optimizers that takes care of the common boilerplate
    code, especially in the constructor.
    """

    def __init__(
        self,
        features: Union[Sequence[Sequence[float]], NDArray[np.float32]],
        classes: Optional[Sequence[int]],
        study_name: str = "",
        add_uuid: bool = True,
        n_trials: int = 10,
        n_cores: int = 1,
        n_seeds: int = 1,
    ) -> None:
        """
        A constructor for the `OptunaOptimizer` class.

        :param features:
            The lists of features of the objects that are used during the training.
        :param classes:
            A list of classes or function values corresponding to the given lists of
            features.
        :param study_name:
            The name of the `optuna` study. If the study with this name is already
            stored in the DB, the tuner will continue the study.
        :param add_uuid:
            Flag for specifying if uuid should be added to the `study_name`.
        :param n_cores:
            The number of cores that `optuna` can use. The (default) value :math:`-1`
            means all of them.
        :param n_trials:
            The number of trials after which the search will be finished.
        :param n_seeds:
            Number of seeds checked per `optuna` trial.
        """
        self._x: Union[Sequence[Sequence[float]], NDArray[np.float32]] = features
        self._y: Optional[Sequence[int]] = classes

        self._study_name: str = study_name

        if add_uuid:
            self._study_name += str(uuid.uuid1())

        self._n_trials: int = n_trials
        self._n_cores: int = n_cores
        self._n_seeds: int = n_seeds

    @staticmethod
    def _get_storage() -> Optional[str]:
        return environ["aqmlator_database_url"]


class ModelFinder(OptunaOptimizer):
    """
    A class for finding the best QNN model for given data and task.
    """

    # TODO TR:  Make it so that the supervised and unsupervised learning parts use
    #           the same pipeline.

    # TODO TR:  Maybe separate class for clustering?

    def __init__(
        self,
        task_type: str,
        features: Union[Sequence[Sequence[float]], NDArray[np.float32]],
        classes: Optional[Sequence[int]] = None,
        device: Optional[qml.Device] = None,
        study_name: str = "QML_Model_Finder_",
        add_uuid: bool = True,
        minimal_accuracy: float = 0.8,
        n_cores: int = -1,
        n_trials: int = 100,
        n_epochs: int = 10,
        n_seeds: int = 5,
        coupling_map: Optional[List[List[int]]] = None,
        d_wave_access: bool = False,
    ):
        """
        A constructor for `ModelFinder` class.

        :param features:
            Features of the objects to be classified. Their order should correspond to
            that of `classes`.
        :param classes:
            Classes of the classified objects. Their order should correspond to that
            of `features`.
        :param study_name:
            The name of the `optuna` study. If the study with this name is already
            stored in the DB, the finder will continue the study.
        :param device:
            The device on which to run the model.
        :param add_uuid:
            Flag for specifying if uuid should be added to the `study_name`.
        :param minimal_accuracy:
            Minimal accuracy after which the training will end.
        :param n_cores:
            The number of cores that `optuna` can use. The (default) value :math:`-1`
            means all of them.
        :param n_trials:
            The number of trials after which the search will be finished.
        :param n_epochs:
            The number of QNN training epochs.
        :param n_seeds:
            Number of seeds checked per `optuna` trial.
        :param coupling_map:
            A list of connected qubits.
        :param d_wave_access:
            Flag for specifying if the model could be run on the D-Wave machine.
        """
        super().__init__(
            features=features,
            classes=classes,
            study_name=study_name,
            add_uuid=add_uuid,
            n_trials=n_trials,
            n_cores=n_cores,
            n_seeds=n_seeds,
        )

        # A dict that will be string models taken under consideration during the
        # model finding.
        self._models_dict: Dict[str, Any] = {}

        self._task_type: str = task_type

        self._n_epochs: int = n_epochs

        self._minimal_accuracy: float = minimal_accuracy

        self._optuna_objective_functions: Dict[
            str, Callable[[optuna.trial.Trial], float]
        ] = {
            MLTaskType.BINARY_CLASSIFICATION: self._simple_model_objective_function,
            MLTaskType.CLASSIFICATION: self._classification_objective_function,
            MLTaskType.REGRESSION: self._simple_model_objective_function,
            MLTaskType.GROUPING: self._grouping_model_objective_function,
        }

        self._optuna_postfix: str = ""

        self.dev: qml.Device = device
        self.device_coupling_map: Optional[List[List[int]]] = coupling_map

        self.d_wave_access: bool = d_wave_access

        try:
            requests.post(
                status_update_endpoint,
                data=json.dumps({self._study_name: "Waiting..."}),
                timeout=1,
            )
        except requests.exceptions.ConnectionError as e:
            print(e)

    def find_model(self) -> None:
        """
        Finds the QNN model that best fits the given data.
        """
        try:
            requests.post(
                status_update_endpoint,
                data=json.dumps({self._study_name: "Tuning..."}),
                timeout=1,
            )
        except requests.exceptions.ConnectionError as e:
            print(e)

        sampler: TPESampler = TPESampler(
            seed=0, multivariate=True, group=True  # For experiments repeatability.
        )

        study: optuna.study.Study = optuna.create_study(
            sampler=sampler,
            study_name=self._study_name,
            load_if_exists=True,
            storage=self._get_storage(),
        )

        study.optimize(
            self._optuna_objective_functions[self._task_type],
            n_trials=self._n_trials,
            n_jobs=self._n_cores,
        )

        try:
            requests.post(
                status_update_endpoint,
                data=json.dumps({self._study_name: "Done."}),
                timeout=1,
            )
        except requests.exceptions.ConnectionError as e:
            print(e)

    def _simple_model_objective_function(self, trial: optuna.trial.Trial) -> float:
        """
        Default objective function of the `optuna` optimizer for the model finding. It
        is meant to work for all the simple (single) models.

        :Note:
            Instead of optimizing the hyperparameters, as `optuna` usually does, this
            optimizes the structure of the VQC for the model.

        :param trial:
            The `optuna` Trial object used to randomize and store the
            results_reconstruction of the optimization.

        :return:
            The average number of calls made to the quantum device (which `optuna`
            wants to minimize).
        """
        self._initialize_model_dict()

        model_type: str = trial.suggest_categorical(
            "model_type" + self._optuna_postfix, list(self._models_dict)
        )

        kwargs: Dict[str, Any] = self._suggest_supervised_model_kwargs(
            trial, model_type
        )

        model: QMLModel = self._models_dict[model_type]["constructor"](**kwargs)

        return self._evaluate_supervised_model(model)

    def _grouping_model_objective_function(self, trial: optuna.trial.Trial) -> float:
        """
        Objective function of the `optuna` optimizer for grouping model finder.

        :param trial:
            The `optuna` Trial object used to randomize and store the values
            used in the optimization.

        :return:
            The average clustering score obtained by the model.
        """
        self._initialize_model_dict()

        model_type: str = trial.suggest_categorical(
            "model_type" + self._optuna_postfix, list(self._models_dict)
        )

        kwargs: Dict[str, Any] = self._suggest_unsupervised_model_kwargs(
            trial, model_type
        )

        model: RBMClustering = self._models_dict[model_type]["constructor"](**kwargs)

        return self._evaluate_unsupervised_model(model)

    def _initialize_model_dict(self) -> None:
        """
        Initializes the models dict used during the model finding.

        :Note:
            The (non-binary) classification task uses binary classification dict to
            create a complex (classification) model.
        """
        self._models_dict = binary_classifiers

        if self._task_type == MLTaskType.REGRESSION:
            self._models_dict = regressors

        if self._task_type == MLTaskType.GROUPING:
            self._models_dict = clustering

    def _evaluate_unsupervised_model(self, model: RBMClustering) -> float:
        """
        Evaluates the performance of the given model. The evaluation is based on the
        Silhouette score (which takes values from [-1, 1]). The higher the score, the
        better the model.

        :param model:
            A `RBMClustering` model to evaluate.

        :return:
            The average Silhouette score obtained by the model.
        """
        score: float = 0
        score_x = np.array(self._x).reshape(
            len(self._x), int(prod(np.array(self._x[0]).shape))
        )

        data: Sequence[Tuple[Tensor, Tensor]] = [
            (Tensor(val), Tensor([-1])) for val in self._x
        ]

        # Type ignore the following line, because Torch isn't type-hinted well enough.
        data_loader: DataLoader[Tuple[Tensor, Tensor]] = DataLoader(
            data  # type: ignore
        )

        for seed in range(self._n_seeds):
            model.rbm.rng = np.random.MT19937(seed)

            model.fit(data_loader)

            labels: List[Tuple[int, ...]] = [
                tuple(model.predict(val[0])) for val in data_loader
            ]
            group_labels: List[Tuple[int, ...]] = list(set(labels))
            groups: List[int] = [group_labels.index(label) for label in labels]

            # Protect against all objects being in the same group. Remember that optuna
            # aim to MINIMIZE the objective function!
            if len(set(groups)) > 1:
                score -= silhouette_score(score_x, groups)
            else:
                score += 1

        return score / self._n_seeds

    def _evaluate_supervised_model(self, model: QMLModel) -> float:
        """
        Evaluates the performance of the given model. The evaluation is based on the
        number of calls to the quantum machine (which are _expensive_) during the
        fitting.

        :param model:
            A `QMLModel` to evaluate.

        The mean number of calls to the quantum machine during the fitting.
        """
        quantum_device_calls: int = 0

        for seed in range(self._n_seeds):
            model.dev = self.dev
            model.seed(seed)

            model.fit(self._x, self._y)

            quantum_device_calls += model.n_executions()

        return quantum_device_calls / self._n_seeds

    def _classification_objective_function(self, trial: optuna.trial.Trial) -> float:
        """
        Objective function of the `optuna` optimizer for classification model finder.

        :param trial:
            The `optuna` Trial object used to randomize and store the
            results_reconstruction of the optimization.

        :return:
            The average number of calls made to the quantum device (which `optuna`
            wants to minimize).
        """
        self._initialize_model_dict()

        n_classes: int = len(np.unique(self._y))

        binary_classifiers_kwargs: List[Dict[str, Any]] = []

        for i in range(n_classes):
            self._optuna_postfix = f"_({i})"

            kwargs: Dict[str, Any] = self._suggest_supervised_model_kwargs(trial, "QNN")

            binary_classifiers_kwargs.append(kwargs)

        qnn_binary_classifiers: List[QNNBinaryClassifier] = []

        for i in range(n_classes):
            qnn_binary_classifiers.append(
                QNNBinaryClassifier(**binary_classifiers_kwargs[i])
            )

        classifier: QNNClassifier = QNNClassifier(
            wires=range(len(self._x)),
            n_classes=n_classes,
            binary_classifiers=qnn_binary_classifiers,
        )

        self._optuna_postfix = ""

        return self._evaluate_supervised_model(classifier)

    def _suggest_supervised_model_kwargs(
        self, trial: optuna.trial.Trial, model_type: str
    ) -> Dict[str, Any]:
        """
        Suggests the kwargs for the specified (qml) model.

        :param trial:
            An optuna trial object for parameters selection.
        :param model_type:
            A string describing the type of the model. Note that the
            `self._models_dict` has to be initialized properly, prior to calling
            this method.

        :return:
            Returns the suggested kwargs for the selected (qml) model type.
        """
        kwargs: Dict[str, Any] = {
            "wires": len(self._x[0]),
            "n_layers": trial.suggest_int(
                "n_layers" + self._optuna_postfix,
                self._models_dict[model_type]["n_layers"][0],
                self._models_dict[model_type]["n_layers"][1],
            ),
            "n_epochs": self._n_epochs,
            "accuracy_threshold": self._minimal_accuracy,
        }

        # TR: Might need to be extended for different arguments type at some point.
        kwargs_data: Dict[str, Any] = self._models_dict[model_type]["kwargs"]

        for kwarg in kwargs_data:
            kwargs[kwarg] = trial.suggest_int(
                kwarg + self._optuna_postfix,
                kwargs_data[kwarg][0],
                kwargs_data[kwarg][1],
            )

        self._suggest_embedding(trial, kwargs)
        self._suggest_layers(trial, kwargs)

        return kwargs

    def _suggest_unsupervised_model_kwargs(
        self, trial: optuna.trial.Trial, model_type: str
    ) -> Dict[str, Any]:
        """
        Suggests the kwargs for the specified (unsupervised learning) model.

        :param trial:
            An optuna trial object for parameters selection.
        :param model_type:
            A string describing the type of the model. Note that the
            `self._models_dict` has to be initialized properly, prior to calling
            this method.

        :return:
            Returns the suggested kwargs for the selected model type.
        """
        # TR: The input shape has to be of form (1, actual_shape).
        lbae_input_shape: Tuple[int] = (1,) + np.array(self._x[0]).shape
        lbae_input_size: int = prod(lbae_input_shape)

        kwargs: Dict[str, Any] = {
            "lbae_input_shape": lbae_input_shape,
            "lbae_out_channels": trial.suggest_int(
                name="lbae_out_channels",
                low=floor(sqrt(lbae_input_size)),
                high=ceil(0.75 * lbae_input_size),
            ),
        }

        kwargs["rbm_n_visible_neurons"] = kwargs["lbae_out_channels"]

        kwargs["rbm_n_hidden_neurons"] = trial.suggest_int(
            name="rbm_n_hidden_neurons",
            low=floor(sqrt(kwargs["lbae_out_channels"])),
            high=ceil(0.75 * kwargs["lbae_out_channels"]),
        )  # Heuristic.

        # TR: Might need to be extended for different arguments type at some point.
        kwargs_data: Dict[str, Any] = self._models_dict[model_type]["kwargs"]

        kwargs["lbae_n_layers"] = trial.suggest_int(
            "lbae_n_layers" + self._optuna_postfix,
            kwargs_data["lbae_n_layers"][0],
            kwargs_data["lbae_n_layers"][1],
        )

        kwargs["fireing_threshold"] = trial.suggest_float(
            "fireing_threshold" + self._optuna_postfix,
            kwargs_data["fireing_threshold"][0],
            kwargs_data["fireing_threshold"][1],
        )

        kwargs["n_epochs"] = self._n_epochs

        return kwargs

    def _suggest_embedding(
        self, trial: optuna.trial.Trial, kwargs: Dict[str, Any]
    ) -> None:
        """
        Using 'optuna', suggest the embedding and its `kwargs`. Everything is then added
        to the given `kwargs`.

        :param trial:
            Optuna `Trial` object that "suggests" the parameters values.
        :param kwargs:
            A dictionary of keyword arguments that will be used to initialize the
            QML model.
        """
        embedding_type: str = trial.suggest_categorical(
            "embedding" + self._optuna_postfix, list(data_embeddings)
        )

        kwargs["embedding_method"] = data_embeddings[embedding_type]["constructor"]

        embedding_kwargs: Dict[str, Any] = {"wires": range(kwargs["wires"])}

        embedding_kwargs.update(data_embeddings[embedding_type]["fixed_kwargs"])

        kwargs["embedding_kwargs"] = embedding_kwargs

    def _suggest_layers(
        self, trial: optuna.trial.Trial, kwargs: Dict[str, Any]
    ) -> None:
        """
        Using `optuna`, suggest the order of layers in the VQC based on the `kwargs`
        given.

        :param trial:
            Optuna `Trial` object that "suggests" the parameters values.
        :param kwargs:
            A dictionary of keyword arguments that will be used to initialize the
            QML model.
        """
        layers: List[Type[qml.operation.Operation]] = []

        for i in range(kwargs["n_layers"]):
            layer_type: str = trial.suggest_categorical(
                f"layer_{i}" + self._optuna_postfix, list(layer_types)
            )
            layers.append(layer_types[layer_type]["constructor"])

            # TR:   So far all the layer types begin with (N_LAYERS, N_WIRES) tuple, and
            #       then proceed with some additional parameters. This may need to be
            #       rethought later.

        kwargs["layers"] = layers
        kwargs.pop("n_layers")

    def __del__(self) -> None:
        try:
            requests.post(
                status_update_endpoint,
                data=json.dumps({self._study_name: "Delete"}),
                timeout=1,
            )
        except requests.exceptions.ConnectionError as e:
            print(e)


class HyperparameterTuner(OptunaOptimizer):
    """
    This class contains the optuna-based tuner for ML Training hyperparameters.

    TODO TR:    Consider renaming this class, as it's only used by the supervised
                learning models (as of now).
    """

    def __init__(
        self,
        features: Sequence[Sequence[float]],
        classes: Sequence[int],
        model: QMLModel,
        study_name: str = "QML_Hyperparameter_Tuner_",
        add_uuid: bool = True,
        n_trials: int = 10,
        n_cores: int = 1,
        n_seeds: int = 1,
    ) -> None:
        """
        A constructor for the `HyperparameterTuner` class.

        :param features:
            The lists of features of the objects that are used during the training.
        :param classes:
            A list of classes corresponding to the given lists of features.
        :param model:
            A model to be trained.
        :param study_name:
            The name of the `optuna` study. If the study with this name is already
            stored in the DB, the tuner will continue the study.
        :param add_uuid:
            Flag for specifying if uuid should be added to the `study_name`.
        :param n_cores:
            The number of cores that `optuna` can use. The (default) value :math:`-1`
            means all of them.
        :param n_trials:
            The number of trials after which the search will be finished.
        :param n_seeds:
            Number of seeds checked per `optuna` trial.
        """
        super().__init__(
            features=features,
            classes=classes,
            study_name=study_name,
            add_uuid=add_uuid,
            n_trials=n_trials,
            n_cores=n_cores,
            n_seeds=n_seeds,
        )

        self._model: QMLModel = model

    def find_hyperparameters(self) -> None:
        """
        Finds the (sub)optimal training hyperparameters.
        """
        sampler: TPESampler = TPESampler(
            seed=0, multivariate=True, group=True  # For experiments repeatability.
        )

        study: optuna.study.Study = optuna.create_study(
            sampler=sampler,
            study_name=self._study_name,
            load_if_exists=True,
            storage=self._get_storage(),
        )

        study.optimize(
            self._optuna_objective, n_trials=self._n_trials, n_jobs=self._n_cores
        )

    @staticmethod
    def _suggest_optimizer(trial: optuna.trial.Trial) -> GradientDescentOptimizer:
        """

        :param trial:
            The `optuna.trial.Trial` object used to randomize and store the
            results_reconstruction of the optimization.
        :return:
            The suggested optimizer.
        """

        optimizer_type: str = trial.suggest_categorical("optimizer", list(optimizers))

        kwargs_data: Dict[str, Any] = optimizers[optimizer_type]["kwargs"]
        kwargs: Dict[str, Any] = {}

        # TR: Might need rebuilding for int and str kwargs.
        for kwarg in kwargs_data:
            kwargs[kwarg] = trial.suggest_float(
                kwarg, kwargs_data[kwarg]["min"], kwargs_data[kwarg]["max"]
            )

        optimizer: GradientDescentOptimizer = optimizers[optimizer_type]["constructor"](
            **kwargs
        )

        return optimizer

    def _optuna_objective(self, trial: optuna.trial.Trial) -> float:
        """
        Objective function of the `optuna` optimizer.

        :param trial:
            The `optuna` Trial object used to randomize and store the
            results_reconstruction of the optimization.
        :return:
        """
        self._model.optimizer = self._suggest_optimizer(trial)

        initial_executions: int
        executions_sum: int = 0

        for _ in range(self._n_seeds):
            initial_executions = self._model.n_executions()
            self._model.weights = np.zeros_like(self._model.weights)

            if self._y is not None:
                self._model.fit(self._x, self._y)

            executions_sum += self._model.n_executions() - initial_executions

        return executions_sum / self._n_seeds



"""
=============================================================================

    This module contains tests for the functionalities in the data_preprocessing module.

=============================================================================

    Copyright 2022 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""
__author__ = "Tomasz Rybotycki"

import unittest
from math import isclose
from typing import List, Sequence, Tuple, Union

from aqmlator.data_acquisition import LearningDatum, SupervisedLearningDatum
from aqmlator.data_preprocessing import (
    LearningDatumMinMaxScaler,
    get_attributes,
    get_targets,
)


class TestDataPreprocessing(unittest.TestCase):
    """
    A `TestCase` class for data_preprocessing module.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """
        csv_learning_data: List[List[float]] = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
        csv_data_targets: List[int] = [1, 2]

        self._learning_data: List[LearningDatum] = [
            LearningDatum(tuple(data)) for data in csv_learning_data
        ]

        self._supervised_learning_data: List[SupervisedLearningDatum] = [
            SupervisedLearningDatum(tuple(csv_learning_data[i]), csv_data_targets[i])
            for i in range(len(csv_data_targets))
        ]

        self._scaler_max: float = 3.14
        self._scaler_min: float = 0

        self._scaler: LearningDatumMinMaxScaler = LearningDatumMinMaxScaler(
            (self._scaler_min, self._scaler_max)
        )

    def test_get_targets(self) -> None:
        """
        Tests `get_targets` method.
        """
        targets: List[Union[str, int, float]] = get_targets(
            self._supervised_learning_data
        )

        self.assertTrue(
            len(targets) == len(self._supervised_learning_data),
            "Number of targets and the data is different.",
        )

        for i in range(len(targets)):
            self.assertTrue(
                targets[i] == self._supervised_learning_data[i].datum_target,
                "Targets are ordered differently than expected.",
            )

    def test_get_attributes_on_learning_data(self) -> None:
        """
        Tests `get_attributes` method on `LearningDatum` objects.
        """
        attributes: List[Tuple[Union[str, float], ...]] = get_attributes(
            self._learning_data
        )

        self.assertTrue(
            len(attributes) == len(self._learning_data),
            "Number of attributes and the learning data is different.",
        )

        for i in range(len(attributes)):
            self.assertTrue(
                attributes[i] == self._learning_data[i].datum_attributes,
                "Attributes are ordered differently than expected from learning data.",
            )

    def test_get_attributes_on_supervised_learning_data(self) -> None:
        """
        Tests `get_attributes` method on `SupervisedLearningDatum` objects.
        """

        supervised_attributes: List[Tuple[Union[str, float], ...]] = get_attributes(
            self._supervised_learning_data
        )

        self.assertTrue(
            len(supervised_attributes) == len(self._supervised_learning_data),
            "Number of attributes and the learning data is different.",
        )

        for i in range(len(supervised_attributes)):
            self.assertTrue(
                supervised_attributes[i]
                == self._supervised_learning_data[i].datum_attributes,
                "Attributes are ordered differently than expected from supervised "
                "learning data.",
            )

    def _learning_datum_min_max_scaler_attribute_tests(
        self, data: Sequence[LearningDatum]
    ) -> None:
        """
        Boilerplate code for testing if the datum attributes are fitted correctly.
        """
        fitted_data: Sequence[LearningDatum] = self._scaler.fit_transform(data)

        self.assertTrue(
            len(fitted_data) == len(data),
            "Fitted data has different len than the original data.",
        )

        # Ensure that the original data wasn't transformed in the process.
        for i in range(len(fitted_data)):
            self.assertFalse(
                fitted_data[i] == data[i],
                "Original data was transformed in the process!",
            )

        # Ensure attributes are within specified bounds.
        min_value: float = min(float(i) for i in fitted_data[0].datum_attributes)
        max_value: float = max(float(i) for i in fitted_data[0].datum_attributes)
        for i in range(1, len(fitted_data)):
            min_value = min(
                tuple(float(i) for i in fitted_data[i].datum_attributes) + (min_value,)
            )
            max_value = max(
                tuple(float(i) for i in fitted_data[i].datum_attributes) + (max_value,)
            )

        self.assertTrue(
            isclose(max_value, self._scaler_max),
            f"Max value is not close to the specified max:"
            f" {max_value} != {self._scaler_max}.",
        )
        self.assertTrue(
            isclose(min_value, self._scaler_min),
            f"Min value is not close to the specified min:"
            f" {min_value} != {self._scaler_min}",
        )

    def test_learning_datum_min_max_scaler_on_learning_datum(self) -> None:
        """
        Tests `LearningDatumMinMaxScaler` on `LearningDatum` objects.
        """
        self._learning_datum_min_max_scaler_attribute_tests(self._learning_data)

    def test_learning_datum_min_max_scaler_on_supervised_learning_datum(self) -> None:
        """
        Tests `LearningDatumMinMaxScaler` on `SupervisedLearningDatum` objects.
        """
        self._learning_datum_min_max_scaler_attribute_tests(
            self._supervised_learning_data
        )

        fitted_data: Sequence[SupervisedLearningDatum] = self._scaler.fit_transform(
            self._supervised_learning_data
        )

        # Check if classes are intact and if the return type is correct.
        for i in range(len(fitted_data)):
            self.assertTrue(
                isinstance(fitted_data[i], SupervisedLearningDatum),
                "Data has wrong type!",
            )
            self.assertTrue(
                self._supervised_learning_data[i].datum_target
                == fitted_data[i].datum_target,
                "Class of the fitted data has changed!",
            )



"""
=============================================================================

    This module contains tests for the functionalities in the data_acquisition module.

=============================================================================

    Copyright 2022 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

import csv
import locale
import os
import unittest
from typing import List, Tuple, Union

from aqmlator.data_acquisition import (
    CSVDataReceiver,
    LearningDatum,
    SupervisedLearningDatum,
)


class TestDataAcquisition(unittest.TestCase):
    """
    A `TestCase` class for data_acquisition module.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """
        csv_learning_data: List[List[float]] = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
        csv_data_targets: List[int] = [1, 2]

        self._learning_data: List[LearningDatum] = [
            LearningDatum(tuple(data)) for data in csv_learning_data
        ]

        self._supervised_learning_data: List[SupervisedLearningDatum] = [
            SupervisedLearningDatum(tuple(csv_learning_data[i]), csv_data_targets[i])
            for i in range(len(csv_data_targets))
        ]

        with open(
            "learning_data.csv", "w", newline="", encoding=locale.getpreferredencoding()
        ) as f:
            writer = csv.writer(f, "excel")
            for data in csv_learning_data:
                writer.writerow(data)

        with open(
            "supervised_learning_data.csv",
            "w",
            newline="",
            encoding=locale.getpreferredencoding(),
        ) as f:
            writer = csv.writer(f, "excel")
            for i in range(len(csv_data_targets)):
                writer.writerow(csv_learning_data[i] + [csv_data_targets[i]])

    def tearDown(self) -> None:
        """
        Ensures that the files created during the tests are deleted.
        """
        self._ensure_deleted("learning_data.csv")
        self._ensure_deleted("supervised_learning_data.csv")

    @staticmethod
    def _ensure_deleted(file_path: str) -> None:
        """
        Ensures that the file on under the given path is deleted.

        :param file_path:
            A path to the file that should be deleted.
        """
        try:
            os.remove(file_path)
        except FileNotFoundError:
            pass

    def test_learning_datum_equality(self) -> None:
        """
        Tests if equality of learning datum was implemented properly.
        """
        x: Tuple[Union[float, str], ...] = (1.23, "lol")
        x2: Tuple[Union[float, str], ...] = (3.21, "nie_lol")
        datum_1: LearningDatum = LearningDatum(x)
        datum_2: LearningDatum = LearningDatum(x)
        datum_3: LearningDatum = LearningDatum(x2)

        self.assertTrue(
            datum_2 == datum_1, "LearningDatum equality implementation has an error."
        )

        self.assertFalse(
            datum_1 == datum_3, "LearningDatum objects shouldn't be equal!"
        )

    def test_supervised_learning_datum_equality(self) -> None:
        """
        Tests if equality of SupervisedLearningDatum was implemented properly.
        """
        x: Tuple[Union[float, str], ...] = (1.23, "lol")
        x2: Tuple[Union[float, str], ...] = (3.21, "nie_lol")
        y: Union[float, str, int] = 1
        datum_1: SupervisedLearningDatum = SupervisedLearningDatum(x, y)
        datum_2: SupervisedLearningDatum = SupervisedLearningDatum(x, y)
        datum_3: SupervisedLearningDatum = SupervisedLearningDatum(x, int(y) + 1)
        datum_4: SupervisedLearningDatum = SupervisedLearningDatum(x2, y)

        self.assertTrue(
            datum_2 == datum_1,
            "SupervisedLearningDatum equality implementation has an error.",
        )

        self.assertFalse(
            datum_1 == datum_3, "SupervisedLearningDatum objects shouldn't be equal!"
        )

        self.assertFalse(
            datum_1 == datum_4, "SupervisedLearningDatum objects shouldn't be equal!"
        )

    def test_csv_learning_data_reading(self) -> None:
        """
        Checks if the (unsupervised) learning data is properly read from the file.
        """
        receiver: CSVDataReceiver = CSVDataReceiver()

        data: List[LearningDatum] = receiver.receive_data("learning_data.csv")

        self.assertTrue(
            len(data) == len(self._learning_data),
            "Received (unsupervised) learning data don't have proper size.",
        )

        for i in range(len(data)):
            self.assertTrue(
                self._learning_data[i] == data[i], "LearningDatum wasn't read properly."
            )

    def test_csv_supervised_data_reading(self) -> None:
        """
        Checks if the supervised learning data is properly read from the file.
        """
        receiver: CSVDataReceiver = CSVDataReceiver(target_index=3)

        data: List[LearningDatum] = receiver.receive_data(
            "supervised_learning_data.csv"
        )

        self.assertTrue(
            len(data) == len(self._supervised_learning_data),
            "Received supervised learning data don't have proper size.",
        )

        for i in range(len(data)):
            self.assertTrue(
                self._supervised_learning_data[i] == data[i],
                "SupervisedLearningDatum wasn't read properly.",
            )



"""
=============================================================================

    This module contains a server implementation for the AQMLator application.

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

from typing import Dict

import uvicorn
from fastapi import FastAPI, Response

app: FastAPI = FastAPI()

status_data: Dict[str, str] = {}

port: str = "8000"
address: str = "127.0.0.1"
status_update_endpoint: str = f"http://{address}:{port}/update_status"


@app.get("/status")
def status() -> Response:
    content: str = ""

    if not status_data:
        content += "No running tuners!"
    else:
        for status_id, tuning_status in status_data.items():
            content += f"Tuner {status_id}: {tuning_status}\n"

    return Response(status_code=200, content=content)


@app.post("/update_status")
async def get_data(data: Dict[str, str]) -> Response:
    for k in data:
        if data[k] == "Delete":
            status_data.pop(k)
        else:
            status_data[k] = data[k]
    return Response(status_code=200, content="Status updated!")


# app.add_api_route("/status", status, methods=["GET"])

if __name__ == "__main__":
    uvicorn.run("server:app", host="127.0.0.1", port=8000, reload=True)



"""
    Machine learning (ML) is one of the fastest-progressing research directions in
    applied computer science. The field investigates the development of algorithms that
    can learn from data by fitting a collection of model parameters to the data by
    iteratively optimizing an objective function. The selection of a model structure, be
    it a neural network topology or kernel function, is a problem-dependent task often
    made by hand. But there exist Auto ML systems that can choose a model automatically
    depending solely on the input data and task.

    Quantum computing (QC) studies how hard computational problems can be efficiently
    solved using quantum mechanics (QM). A large-scale error-corrected quantum computer
    can solve computational problems that don't have a classical solution. A prime
    example of that is Shor's algorithm for integer factorization. The
    'holy grail' of applied QC is the so-called quantum supremacy or quantum
    advantage. That is the name for a technological milestone that would mark the moment
    when quantum machines will solve a specific task faster than the most advanced
    supercomputer. Although there have already been several quantum supremacy claims in
    recent years, there are no practical problems solved using quantum computing yet.

    The search for such practical problems focuses on applications in the soft computing
    areas that are less susceptible to current quantum hardware imperfections. One of
    the possible applications of QC is quantum machine learning (QML). Quantum computers
    can be employed to build ML models that can be fit to data and then used during the
    inference process. In one of the QML scenarios, a variational quantum circuit
    forming a quantum neural network (QNN) constitutes only one part of the ML data
    processing pipeline. Since designing such a pipeline with a quantum component is
    challenging for non-experts in QC, we propose an auto ML solution that uses QML
    techniques.

    We call our solution (A)uto (Q)uantum (M)achine (L)earning platform---AQMLator. It
    aims to enable data scientists, especially those without knowledge of QC, to use
    QML models. Given the data, the system will seek the model that fits the
    data best. Additionally, it will propose the circuit structure of the
    VQC model and its weights. Ultimately, the user can use the model as-is
    or encompass it in a hybrid model e.g. by extracting the proposed VQC model as a
    torch layer.
"""



"""
=============================================================================

    This module contains tests for the functionalities in the database module.

=============================================================================

    Copyright 2022 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

import os
import unittest

from aqmlator.database import export_data_to_sqlite_database


class TestDatabaseFunctionalities(unittest.TestCase):
    """
    A `TestCase` class for the functionalities in the `database` module.
    """

    def setUp(self) -> None:
        """
        Sets up the class for the tests.
        """
        self.sqlite_db_name: str = "aqmlatorSQLite.db"

    def tearDown(self) -> None:
        """
        Cleans up after tests.
        """
        if os.path.exists(self.sqlite_db_name):
            os.remove(self.sqlite_db_name)

    def test_exporting_to_sqlite_database(self) -> None:
        """
        Checks if the SQLite database is properly created.
        """
        export_data_to_sqlite_database(self.sqlite_db_name)
        self.assertTrue(os.path.exists(self.sqlite_db_name))



"""
=============================================================================

    This module contains the functionalities related to the quantum machine learning.

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"

import abc
import random
from itertools import chain
from math import prod
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pennylane as qml
import torch
from dimod.core.sampler import Sampler
from lightning.pytorch.trainer import Trainer
from numpy.typing import NDArray
from pennylane import numpy as np
from pennylane.kernels import target_alignment
from pennylane.optimize import GradientDescentOptimizer, NesterovMomentumOptimizer
from pennylane.templates.embeddings import AmplitudeEmbedding, AngleEmbedding
from pennylane.templates.layers import StronglyEntanglingLayers
from pennylane.transforms import transpile
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC
from torch import Tensor
from torch.utils.data import DataLoader

from qbm4eo.encoder import LBAEEncoder
from qbm4eo.lbae import LBAE
from qbm4eo.rbm import RBM, AnnealingRBMTrainer, CD1Trainer, RBMTrainer

ModelOutput = TypeVar("ModelOutput", float, int)


class QMLModel(abc.ABC):
    """
    A boilerplate class, providing an interface for future QML models.
    """

    def __init__(
        self,
        wires: Union[int, Sequence[int]],
        device: Optional[qml.Device] = None,
        optimizer: Optional[GradientDescentOptimizer] = None,
        embedding_method: Optional[Type[qml.operation.Operation]] = None,
        embedding_kwargs: Optional[Dict[str, Any]] = None,
        layers: Optional[Sequence[Type[qml.operation.Operation]]] = None,
        validation_set_size: float = 0.2,
        rng_seed: int = 42,
        coupling_map: Optional[Sequence[Sequence[int]]] = None,
        n_qubit: Optional[int] = None,
    ) -> None:
        """
        The constructor for the `QMLModel` class.

        :param wires:
            The wires to use in the VQC or the number of qubits (and wires) used in the
            VQC.
        :param device:
            A device on which the model should operate.
        :param optimizer:
            The optimizer that will be used in the training. `NesterovMomentumOptimizer`
            with default parameters will be used as default.
        :param embedding_method:
            Embedding method of the data. By default - when `None` is specified - the
            QMLModel will use `AngleEmbedding`. See `_prepare_default_embedding`
            for parameters details.
        :param embedding_kwargs:
            Keyword arguments for the embedding method. If `None` are specified the
            QMLModel will use the default embedding method.
        :param layers:
            Layers to be used in the VQC. The layers will be applied in the given order.
            A double `StronglyEntanglingLayer` will be used if `None` is given.
        :param validation_set_size:
            A part of the training set that will be used for QMLModel validation.
            It should be from (0, 1).
        :param rng_seed:
            A seed used for random weights initialization.
        :param coupling_map:
            A description of connections between the qubits in the device.
        """
        self.dev: qml.Device = device

        self.wires: Sequence[int]

        if isinstance(wires, int):
            self.wires = list(range(wires))
        else:
            self.wires = wires

        self._training_X: Sequence[Sequence[float]]
        self._training_y: Sequence[int]
        self._validation_X: Sequence[Sequence[float]]
        self._validation_y: Sequence[int]

        self._embedding_method: Type[qml.operation.Operation]
        self._embedding_kwargs: Dict[str, Any]

        if embedding_method is None or embedding_kwargs is None:
            self._prepare_default_embedding()
        else:
            self._embedding_method = embedding_method
            self._embedding_kwargs = embedding_kwargs

        self._layers: Sequence[Type[qml.operation.Operation]]

        if layers is None:
            self._prepare_default_layers()
        else:
            self._layers = layers

        self._rng_seed: int = rng_seed
        self._rng: random.Random = random.Random(rng_seed)

        self._validation_set_size: float = validation_set_size

        if optimizer is None:
            optimizer = NesterovMomentumOptimizer()

        self.optimizer: GradientDescentOptimizer = optimizer
        self.weights: Sequence[float]

        self.coupling_map: Optional[Sequence[Sequence[int]]] = coupling_map

        if not n_qubit:
            n_qubit = len(self.wires)

        self.n_qubit: int = n_qubit

    def seed(self, new_seed: int) -> None:
        """
        Sets up the new seed.

        :param new_seed:
            New seed to be applied to the model.
        """
        self._rng_seed = new_seed

    def n_executions(self) -> int:
        """
        Returns number of VQC executions so far.

        :return:
            Returns the number of times the quantum device was called.
        """
        return self.dev.num_executions

    @abc.abstractmethod
    def fit(
        self,
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Optional[Sequence[ModelOutput]],
    ) -> "QMLModel":
        """
        The model training method.

        :param X:
            The lists of features of the objects that are used during the training.
        :param y:
            A list of classes corresponding to the given lists of features.

        :raise NotImplementedError:
            If the method is not implemented.

        :return:
            Returns self after training.
        """
        raise NotImplementedError

    def _prepare_default_embedding(self) -> None:
        """
        Prepares the default embedding method is `None` was specified or if the
        kwargs were `None`. The default one is simple `AngleEmbedding`.
        """
        self._embedding_method = AngleEmbedding
        self._embedding_kwargs = {"wires": self.wires}

    def _prepare_default_layers(self) -> None:
        """
        Prepares the default layers of the model if `None` was given in either
        `layers` or `layers_weights_number` arguments of the constructor. We will
        use a double strongly entangling layer.
        """
        self._layers = [StronglyEntanglingLayers] * 2

    def _split_data_for_training(
        self,
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Sequence[ModelOutput],
    ) -> None:
        """
        Splits the objects into validation and training sets. Should be called before
        the training, usually near the beginning of the `fit` method.

        :param X:
            All object features that will be split into training and validation set.
        :param y:
            Corresponding classes that will be split into training and validation set.
        """
        (
            self._training_X,
            self._validation_X,
            self._training_y,
            self._validation_y,
        ) = train_test_split(
            X,
            y,
            test_size=self._validation_set_size,
            random_state=self._rng_seed,
        )


class QNNModel(QMLModel, abc.ABC):
    """
    A boilerplate class, providing an interface for future QNN-Based models.
    """

    def __init__(
        self,
        wires: Union[int, Sequence[int]],
        batch_size: int,
        n_epochs: int = 1,
        device: Optional[qml.Device] = None,
        optimizer: Optional[GradientDescentOptimizer] = None,
        embedding_method: Optional[Type[qml.operation.Operation]] = None,
        embedding_kwargs: Optional[Dict[str, Any]] = None,
        layers: Optional[Sequence[Type[qml.operation.Operation]]] = None,
        accuracy_threshold: float = 0.8,
        initial_weights: Optional[Sequence[float]] = None,
        rng_seed: int = 42,
        validation_set_size: float = 0.2,
        prediction_function: Optional[
            Callable[
                [Sequence[Sequence[float]]],
                Sequence[ModelOutput],
            ]
        ] = None,
        debug_flag: bool = True,
        coupling_map: Optional[Sequence[Sequence[int]]] = None,
        n_qubit: Optional[int] = None,
    ) -> None:
        """
        The constructor for the `QNNModel` class.

        :param wires:
            The wires to use in the VQC or the number of qubits (and wires) used in the
            VQC.
        :param batch_size:
            Size of a batches used during the training.
        :param n_epochs:
            The number of training epochs.
        :param device:
            A device on which the model should operate.
        :param optimizer:
            The optimizer that will be used in the training. `NesterovMomentumOptimizer`
            with default parameters will be used as default.
        :param embedding_method:
            Embedding method of the data. By default - when `None` is specified - the
            model will use `AmplitudeEmbedding`. See `_prepare_default_embedding`
            for parameters details.
        :param embedding_kwargs:
            Keyword arguments for the embedding method. If none are specified the
            model will use the default embedding method.
        :param layers:
            Layers to be used in the VQC. The layers will be applied in the given order.
            A double `StronglyEntanglingLayer` will be used if `None` is given.
        :param accuracy_threshold:
            The satisfactory accuracy of the model.
        :param initial_weights:
            The initial weights for the training.
        :param rng_seed:
            A seed used for random weights initialization.
        :param validation_set_size:
            A part of the training set that will be used for model validation.
            It should be from (0, 1).
        :param prediction_function:
            A prediction function that will be used to process the output of the VQC.
            If `None` then the default one (for given model) will be used.
        :param debug_flag:
            A flag informing the model if the training info should be printed to the
            console or not.
        :param coupling_map:
            A description of connections between the qubits in the device.
        """

        super().__init__(
            wires,
            device,
            optimizer,
            embedding_method,
            embedding_kwargs,
            layers,
            validation_set_size,
            rng_seed,
            coupling_map,
            n_qubit,
        )

        self._n_epochs: int = n_epochs
        self._batch_size: int = batch_size

        self.accuracy_threshold: float = accuracy_threshold

        self._weights_length: int = 0

        for layer in self._layers:
            self._weights_length += prod(
                layer.shape(n_layers=1, n_wires=len(self.wires))
            )

        if initial_weights is None or len(initial_weights) != self._weights_length:
            initial_weights = [
                np.pi * self._rng.random() for _ in range(self._weights_length)
            ]

        self.weights = initial_weights

        self.circuit: Optional[qml.QNode] = None

        self._debug_flag: bool = debug_flag

        if prediction_function is None:
            prediction_function = self._default_prediction_function

        self._prediction_function: Callable[
            [Sequence[Sequence[float]]], Union[Sequence[int], Sequence[float]]
        ] = prediction_function

    def _create_circuit(self, interface: str = "autograd") -> qml.QNode:
        def circuit(
            inputs: Union[Sequence[float], torch.Tensor],
            weights: Union[np.ndarray, torch.Tensor],
        ) -> Sequence[float]:
            """
            Returns the expectation value of the first qubit of the VQC of which the
            weights are optimized during the learning process.

            :param inputs:
                Feature vector representing the object for which value is being
                predicted.

                :note:
                This argument needs to be named `inputs` for torch to be able to use
                the `circuit` method.
            :param weights:
                Weights that will be optimized during the learning process.

            :return:
                The expectation value (from range [-1, 1]) of the measurement in the
                computational basis of given circuit.
            """

            if isinstance(inputs, torch.Tensor):
                inputs = self._prepare_torch_inputs(inputs)

            with qml.QueuingManager.stop_recording():
                ops = self._embedding_method(inputs, **self._embedding_kwargs).expand()

            for op in ops:
                qml.apply(op)

            start_weights: int = 0

            for i, layer in enumerate(self._layers):
                layer_shape: Tuple[int, ...] = layer.shape(
                    n_layers=1, n_wires=len(self.wires)
                )

                layer_weights = weights[
                    start_weights : start_weights + prod(layer_shape)
                ]

                start_weights += prod(layer_shape)

                layer_weights = layer_weights.reshape(layer_shape)

                with qml.QueuingManager.stop_recording():
                    ops = layer(layer_weights, wires=self.wires).expand()

                for op in ops:
                    qml.apply(op)

            return [qml.expval(qml.PauliZ((i))) for i in self.wires]

        if self.coupling_map:
            circuit = transpile(coupling_map=self.coupling_map)(circuit)

        return qml.QNode(circuit, self.dev, interface=interface)

    def get_circuit_expectation_values(
        self, features_lists: Sequence[Sequence[float]]
    ) -> np.ndarray:
        """
        Computes and returns the expectation value of the `PauliZ` measurement of the
        first qubit of the VQC.

        :param features_lists:
            Features that will be encoded at the input of the circuit.

        :return:
            The expectation value of the `PauliZ` measurement on the first qubit of the
            VQC.
        """

        if not self.circuit:
            self.circuit = self._create_circuit()

        expectation_values: List[Sequence[float]] = []

        for features in features_lists:
            expectation_values.append(self.circuit(features, np.array(self.weights)))

        return np.array(expectation_values)

    @abc.abstractmethod
    def _cost(
        self,
        weights: Sequence[float],
        X: Sequence[Sequence[float]],
        y: Sequence[int],
    ) -> float:
        """
        Evaluates and returns the cost function value for the training purposes.

        :param X:
            The lists of features of the objects for which values are being predicted
            during the training.

        :param y:
            Outputs corresponding to the given features.

        :raise NotImplementedError:
            If the method is not implemented.

        :return:
            The value of the square loss function.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _default_prediction_function(
        self, circuit_outputs: Sequence[Sequence[float]]
    ) -> Sequence[ModelOutput]:
        """
        The default prediction function that should be specified for every QNN-based
        model.

        :param circuit_outputs:
            The output of the VQC.

        :return:
            Returns prediction value for the given problem.
        """
        pass

    def fit(
        self,
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Optional[Sequence[ModelOutput]],
    ) -> "QNNModel":
        """
        The model training method.

        TODO TR: How to break this method down into smaller ones?

        :param X:
            The lists of features of the objects that are used during the training.
        :param y:
            A list of outputs corresponding to the given lists of features.

        :raise AttributeError:
            If the device is not specified.

        :raise AttributeError:
            If the `y` is not specified.

        :return:
            Returns `self` after training.
        """

        if not self.dev:
            raise AttributeError("Specify the device (dev) before fitting.")

        if y is None:
            raise AttributeError("Missing y in supervised learning model.")

        self.circuit = self._create_circuit()

        self._split_data_for_training(X, y)

        n_batches: int = max(1, len(X) // self._batch_size)

        feature_batches = np.array_split(np.arange(len(self._training_X)), n_batches)

        best_weights: Sequence[float] = self.weights
        best_accuracy: float = self.score(self._validation_X, self._validation_y)

        self.weights = np.array(self.weights, requires_grad=True)
        cost: float
        batch_indices: np.tensor  # Of ints.

        def _batch_cost(weights: Sequence[float]) -> float:
            """
            The cost function evaluated on the training data batch.

            :param weights:
                The weights to be applied to the VQC.

            :return:
                The value of `self._cost` function evaluated of the training data
                batch.
            """
            return self._cost(
                weights,
                self._training_X[batch_indices],
                self._training_y[batch_indices],
            )

        for it, batch_indices in enumerate(
            chain(*(self._n_epochs * [feature_batches]))
        ):
            # Update the weights by one optimizer step
            self.weights, cost = self.optimizer.step_and_cost(_batch_cost, self.weights)

            # Compute accuracy on the validation set
            accuracy_validation = self.score(self._validation_X, self._validation_y)

            # Make decision about stopping the training basing on the validation score
            if accuracy_validation >= best_accuracy:
                best_accuracy = accuracy_validation
                best_weights = self.weights

            if self._debug_flag:
                print(
                    f"It: {it + 1} / {self._n_epochs * n_batches} | Cost: {cost} |"
                    f" Accuracy (validation): {accuracy_validation}"
                )

            if accuracy_validation >= self.accuracy_threshold:
                break

        self.weights = best_weights

        # TR: Required for proper serialization.
        # TODO TR:  Maybe there's a better way to do it
        self.circuit = None

        return self

    @abc.abstractmethod
    def score(
        self,
        X: Sequence[Sequence[float]],
        y: Sequence[Union[int, float]],
        sample_weight: Sequence[float] = None,
    ) -> float:
        """
        Computes and returns score of the model.

        :note:
            There are no typehints, same as in sklearn `score` functions that we will
            use.

        :param X:
            Test samples.
        :param y:
            True labels for `X`.
        :param sample_weight:
            Sample weights.

        :raise NotImplementedError:
            If the method is not implemented.

        :return:
            Mean accuracy of `self.predict(X)` w.r.t. `y`.
        """
        raise NotImplementedError

    def _prepare_torch_inputs(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Torch inputs might need some manual preprocessing in some cases. This method
        handles this preprocessing.

        :param inputs:
            Inputs given by the
        :return:
            Preprocessed inputs.
        """
        # TODO TR: Think of better way to do it.
        if self._embedding_method == AmplitudeEmbedding:
            padding: torch.Tensor = torch.zeros([2 ** len(self.wires) - len(inputs)])
            inputs = torch.cat((inputs, padding))

        return inputs

    def get_torch_layer(self) -> torch.nn.Module:
        """
        This method creates a PyTorch (quantum) layer based on the VQC.

        :return:
            Returns a PyTorch Layer made from the VQC.
        """

        weight_shapes: Dict[str, int] = {"weights": len(self.weights)}
        return qml.qnn.TorchLayer(self._create_circuit("torch"), weight_shapes)

    def predict(
        self, features: Sequence[Sequence[float]]
    ) -> Union[Sequence[float], Sequence[int]]:
        """
        Returns predictions of the model for the given features.

        :param features:
            Features of the objects for which the model will predict the values.
        :return:
            Values predicted for given features.
        """

        self.circuit = self._create_circuit()

        results: Union[Sequence[int], Sequence[float]] = self._prediction_function(
            self.get_circuit_expectation_values(features)
        )

        # TR: Required for proper serialization.
        # TODO TR:  Maybe there's a better way to do it
        self.circuit = None

        return results


class QNNBinaryClassifier(ClassifierMixin, QNNModel):
    """
    This class implements a binary classifier that uses Quantum Neural Networks.

    The classifier expects two classes {0, 1}.
    """

    def _cost(
        self,
        weights: Sequence[float],
        X: Sequence[Sequence[float]],
        y: Sequence[int],
    ) -> float:
        """
        Evaluates and returns the cost function value for the training purposes.

        :param X:
            The lists of features of the objects that are being classified during the
            training.

        :param y:
            Classes corresponding to the given features.

        :return:
            The value of the square loss function.
        """

        self.circuit = self._create_circuit()

        expectation_values: np.ndarray = np.array(
            [self.circuit(x, weights)[0] for x in X]
        )

        return np.mean((expectation_values - np.array(y)) ** 2)

    def _default_prediction_function(
        self, circuit_outputs: Sequence[Sequence[float]]
    ) -> Sequence[ModelOutput]:
        """
        The default prediction function of the QNNClassifier.

        :param circuit_outputs:
            The outputs of the VQC.

        :return:
            Returns classification prediction value for the given problem.
        """
        return [2 * int(val >= 0.0) - 1 for val in [x[0] for x in circuit_outputs]]


class QNNLinearRegression(RegressorMixin, QNNModel):
    """
    This class implements a linear regressor that uses Quantum Neural Networks.
    """

    def _cost(
        self,
        weights: Sequence[float],
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Sequence[int],
    ) -> float:
        """
        Evaluates and returns the cost function value for the training purposes.

        :param X:
            The lists of features of the objects that are being classified during the
            training.

        :param y:
            Classes corresponding to the given features.

        :return:
            The value of the square loss function.
        """
        self.circuit = self._create_circuit()

        expected_values = [self.circuit(x, weights) for x in X]

        predicted_values: np.ndarray = np.array(
            [
                sum(np.log(((i + 1) / 2) / (1 - ((i + 1) / 2))) for i in x)
                for x in expected_values
            ]
        )

        return np.mean(predicted_values - np.array(y) ** 2)

    def _default_prediction_function(
        self, circuit_outputs: Sequence[Sequence[float]]
    ) -> Sequence[ModelOutput]:
        """
        The default prediction function of the QNNClassifier.

        :param circuit_outputs:
            The outputs of the VQC.

        :return:
            Returns classification prediction value for the given problem.
        """
        predicted_values: List[ModelOutput] = [
            sum(np.log(((i + 1) / 2) / (1 - ((i + 1) / 2))) for i in x)
            for x in circuit_outputs
        ]

        return predicted_values


class QuantumKernelBinaryClassifier(QMLModel, ClassifierMixin):
    """
    This class implements the binary classifier based on quantum kernels.
    """

    def __init__(
        self,
        wires: Union[int, Sequence[int]],
        n_epochs: int = 10,
        kta_subset_size: int = 5,
        device: Optional[qml.Device] = None,
        optimizer: Optional[GradientDescentOptimizer] = None,
        embedding_method: Optional[Type[qml.operation.Operation]] = None,
        embedding_kwargs: Optional[Dict[str, Any]] = None,
        layers: Optional[Sequence[Type[qml.operation.Operation]]] = None,
        initial_weights: Optional[Sequence[float]] = None,
        rng_seed: int = 42,
        accuracy_threshold: float = 0.8,
        validation_set_size: float = 0.2,
        debug_flag: bool = True,
        coupling_map: Optional[Sequence[Sequence[int]]] = None,
    ) -> None:
        """
        A constructor for the `QuantumKernelBinaryClassifier` class.

        :param wires:
            The wires to use in the VQC or the number of qubits (and wires) used in the
            VQC.
        :param n_epochs:
            The maximal number of training iterations.
        :param kta_subset_size:
            The number of objects used to evaluate the kernel target alignment method
            in the cost function.
        :param device:
            A device on which the model should operate.
        :param optimizer:
            The optimizer that will be used in the training. `NesterovMomentumOptimizer`
            with default parameters will be used as default.
        :param embedding_method:
            Embedding method of the data. By default - when `None` is specified - the
            classifier will use `AngleEmbedding`. See `_prepare_default_embedding`
            for parameters details.
        :param embedding_kwargs:
            Keyword arguments for the embedding method. If none are specified the
            classifier will use the default embedding method.
        :param layers:
            A list of `layer` functions to be applied in the kernel ansatz VQC.
        :param initial_weights:
            The weights using which the training will start.
        :param rng_seed:
            A random seed used to initialize the weights (if no weights are given).
        :param accuracy_threshold:
            The accuracy after which the training is considered complete.
        :param validation_set_size:
            A part of the training set that will be used for classifier validation.
            It should be from (0, 1).
        :param debug_flag:
            A flag informing the classifier if the training info should be printed
            to the console or not.
        :param coupling_map:
            A description of connections between the qubits in the device.
        """
        super().__init__(
            wires,
            device,
            optimizer,
            embedding_method,
            embedding_kwargs,
            layers,
            validation_set_size,
            rng_seed,
            coupling_map,
        )

        self.n_epochs: int = n_epochs

        self._kta_subset_size: int = kta_subset_size

        self._accuracy_threshold = accuracy_threshold
        self._validation_set_size = validation_set_size

        self._weights_length: int = 0

        for layer in self._layers:
            self._weights_length += prod(
                layer.shape(n_layers=1, n_wires=len(self.wires))
            )

        self._rng_seed: int = rng_seed
        self._rng: random.Random = random.Random(rng_seed)

        if initial_weights is None or len(initial_weights) != self._weights_length:
            initial_weights = [
                np.pi * self._rng.random() for _ in range(self._weights_length)
            ]

        self.weights = np.array(initial_weights, requires_grad=True)

        self._debug_flag: bool = debug_flag

        self._classifier: SVC = SVC()

    def _ansatz(self, weights: Sequence[float], features: Sequence[float]) -> None:
        """
        A VQC ansatz that will be used in defining the quantum kernel function.

        :note:
            Templates expanding is necessary for the transpiling!

        :param weights:
            Weights that will be optimized during the learning process.
        :param features:
            Feature vector representing the object that is being classified.
        """

        start_weights: int = 0

        for layer in self._layers:
            with qml.QueuingManager.stop_recording():
                ops = self._embedding_method(
                    features, **self._embedding_kwargs
                ).expand()

            for op in ops:
                qml.apply(op)

            layer_shape: Tuple[int, ...] = layer.shape(
                n_layers=1, n_wires=len(self.wires)
            )

            layer_weights = weights[start_weights : start_weights + prod(layer_shape)]
            start_weights += prod(layer_shape)
            layer_weights = np.array(layer_weights).reshape(layer_shape)

            with qml.QueuingManager.stop_recording():
                ops = layer(layer_weights, wires=self.wires).expand()

            for op in ops:
                qml.apply(op)

    def _create_transform(
        self,
    ) -> qml.QNode:
        """
        Creates a feature map VQC based on the current kernel.

        :return:
            The feature map VQC based on the current kernel.
        """

        # @qml.qnode(self.dev)
        def transform(
            weights: Sequence[float], features: Sequence[float]
        ) -> List[qml.measurements.ExpectationMP]:
            """
            The definition of the feature map VQC.

            :param weights:
                Parameters of the VQC.
            :param features:
                The features of the object to be transformed.

            :return:
                The result of `qml.PauliZ` measurements on the feature map VQC.
            """
            self._ansatz(weights, features)

            # TODO TR: Is this a good measurement to return?
            return [qml.expval(qml.PauliZ((i,))) for i in self.wires]

        if self.coupling_map:
            transform = transpile(coupling_map=self.coupling_map)(transform)

        return qml.QNode(transform, self.dev)

    def _create_kernel(
        self,
    ) -> Callable[[Sequence[float], Sequence[float], Sequence[float]], float]:
        """
        Prepares the VQC that will return the quantum kernel value for given data
        points.

        :return:
            The VQC structure representing the kernel function.
        """

        # Adjoint circuits is prepared pretty easily.
        adjoint_ansatz: Callable[
            [Sequence[float], Sequence[float]], None
        ] = qml.adjoint(self._ansatz)

        # @qml.qnode(self.dev)
        def kernel_circuit(
            weights: Sequence[float],
            first_features: Sequence[float],
            second_features: Sequence[float],
        ) -> qml.measurements.ProbabilityMP:
            """
            The VQC returning the quantum embedding kernel circuit based on given
            ansatz.

            :param weights:
                Weights to be applied to the VQC ansatz.
            :param first_features:
                Features of the first object.
            :param second_features:
                Features of the second objects.

            :return:
                The probability of observing respective computational-base states.
            """
            self._ansatz(weights, first_features)
            adjoint_ansatz(weights, second_features)
            return qml.probs(wires=self.wires)

        if self.coupling_map:
            kernel_circuit = transpile(coupling_map=self.coupling_map)(kernel_circuit)

        kernel_circuit = qml.QNode(kernel_circuit, device=self.dev)

        def kernel(
            weights: Sequence[float],
            first_features: Sequence[float],
            second_features: Sequence[float],
        ) -> float:
            """
            A method representing the quantum embedding kernel based on the given
            ansatz.

            :param weights:
                Weights to be applied to the VQC ansatz.
            :param first_features:
                Features of the first object.
            :param second_features:
                Features of the second objects.

            :return:
                The value of the kernel (or of measuring the zero state after running
                the VQC).
            """

            # The `np.array` casting is required so that indexing is "legal".
            return np.array(kernel_circuit(weights, first_features, second_features))[0]

        return kernel

    def _kernel_matrix_function(
        self, x: Sequence[Sequence[float]], y: Sequence[int]
    ) -> Callable[[Sequence[float], Sequence[float]], float]:
        """
        Prepares and returns the `kernel_matrix` function that uses the
        trained kernel.

        :param x:
            The lists of features of the objects that are used during the
            training.
        :param y:
            The classes corresponding to the given features.

        :return:
            The `kernel_matrix` function that uses the trained kernel.
        """
        kernel: Callable[
            [Sequence[float], Sequence[float], Sequence[float]], float
        ] = self._create_kernel()

        return qml.kernels.kernel_matrix(
            list(x),
            list(y),
            lambda x1, x2: kernel(self.weights, x1, x2),
        )

    def fit(
        self,
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Optional[Sequence[ModelOutput]],
    ) -> "QuantumKernelBinaryClassifier":
        """
        The classifier training method.

        TODO TR: How to break this method down into smaller ones?

        :param X:
            The lists of features of the objects that are used during the training.
        :param y:
            A list of classes corresponding to the given lists of features. The classes
            should be from set {-1, 1}.

        :raise AttributeError:
            If the device is not specified.

        :raise AttributeError:
            If the `y` is not specified.

        :return:
            Returns `self` after training.
        """

        if not self.dev:
            raise AttributeError("Specify the device (dev) before fitting.")

        if y is None:
            raise AttributeError("Missing y in supervised learning model.")

        kernel: Callable[
            [Sequence[float], Sequence[float], Sequence[float]], float
        ] = self._create_kernel()

        self._split_data_for_training(X, y)

        def cost(weights: Sequence[float]) -> float:
            """
            A cost function used during the learning. We will use negative of kernel
            target alignment method, as the `pennylane` optimizers are meant for
            minimizing the objective function and `target_alignment` function is a
            similarity measure between the kernels.

            :param weights:
                Weights to be applied into kernel VQC.

            :return:
                The negative value of KTA for given weights.
            """

            # Choose subset of datapoints to compute the KTA on.
            subset: np.ndarray = np.array(
                self._rng.sample(
                    list(range(len(self._training_X))), self._kta_subset_size
                )
            )

            return -target_alignment(
                list(self._training_X[subset]),
                list(self._training_y[subset]),
                lambda x1, x2: self._create_kernel()(weights, x1, x2),
                assume_normalized_kernel=True,
            )

        for i in range(self.n_epochs):
            self.weights, cost_val = self.optimizer.step_and_cost(cost, self.weights)

            current_alignment = target_alignment(
                list(X),
                list(y),
                lambda x1, x2: kernel(self.weights, x1, x2),
                assume_normalized_kernel=True,
            )

            # Second create a kernel matrix function using the trained kernel.
            self._classifier = SVC(kernel=self._kernel_matrix_function).fit(X, y)

            accuracy: float = self.score(self._validation_X, self._validation_y)

            if self._debug_flag:
                print(
                    f"Step {i + 1}: "
                    f"Alignment = {current_alignment:.3f} | "
                    f"Step cost value = {cost_val} | "
                    f"Validation Accuracy {accuracy:.3f}"
                )
                print(self.weights)

            if accuracy >= self._accuracy_threshold:
                break

        return self

    def transform(
        self, features_lists: Sequence[Sequence[float]]
    ) -> List[List[qml.measurements.ExpectationMP]]:
        """
        Maps the object described by the `features` into it's representation in the
        feature space.

        :param features_lists:
            The features of the object to be mapped.

        :return:
            The representation of the given object in the feature space.
        """
        transform: qml.QNode = self._create_transform()

        mapped_features: List[List[qml.measurements.ExpectationMP]] = []

        for features_list in features_lists:
            mapped_features.append(transform(self.weights, features_list))

        return mapped_features

    def predict(
        self, features_lists: Sequence[Sequence[float]]
    ) -> Sequence[ModelOutput]:
        """
        Predicts and returns the classes of the objects for which features were given.
        It applies current `self.weights` as the parameters of VQC.

        :param features_lists:
            Objects' features to be encoded at the input of the VQC.

        :return:
            The results_reconstruction - classes 0 or 1 - of the classification.
            The data structure of the returned object is `np.ndarray` with `dtype=bool`.
        """
        return self._classifier.predict(features_lists)


class QNNClassifier(QMLModel, ClassifierMixin):
    """
    This class implements a quantum classifier based on the multiple binary quantum
    classifiers.
    """

    def __init__(
        self,
        wires: Union[int, Sequence[int]],
        n_classes: int,
        binary_classifiers: Optional[Sequence[QNNBinaryClassifier]] = None,
        batch_size: int = 10,
        accuracy_threshold: float = 0.8,
        device: Optional[qml.Device] = None,
        optimizer: Optional[GradientDescentOptimizer] = None,
        embedding_method: Optional[Type[qml.operation.Operation]] = None,
        embedding_kwargs: Optional[Dict[str, Any]] = None,
        layers: Optional[Sequence[Type[qml.operation.Operation]]] = None,
        validation_set_size: float = 0.2,
        rng_seed: int = 42,
    ) -> None:
        """
        The constructor for the `QuantumClassifier` class.

        :param wires:
            The wires to use in the VQC or the number of qubits (and wires) used in the
            VQC. It will be used in the `qml.Device` specification.
        :param n_classes:
            The number of classes in the classification task.
        :param binary_classifiers:
            Binary classifiers that will be used in the classification. If `None` is
            given, then the `QuantumClassifier` will produce default binary classifiers.
        :param batch_size:
            Batch size using during binary classifiers fitting.
        :param accuracy_threshold:
            The target minimal accuracy of the classifier. Note that it reflects total
            accuracy of the classifier, which is lower than accuracy of each binary
            classifier.
        :param device:
            A device on which the model should operate.
        :param optimizer:
            The optimizer that will be used in the training. `NesterovMomentumOptimizer`
            with default parameters will be used as default. It will be used in each of
            the binary classifiers initialized by `QuantumClassifier`.
        :param embedding_method:
            Embedding method of the data. By default - when `None` is specified - the
            QMLModel will use `AngleEmbedding`. See `_prepare_default_embedding`
            for parameters details. It will be used in each of the binary classifiers
            initialized by `QuantumClassifier`.
        :param embedding_kwargs:
            Keyword arguments for the embedding method. If `None` are specified the
            QMLModel will use the default embedding method. It will be used in each of
            the binary classifiers initialized by `QuantumClassifier`.
        :param layers:
            Layers to be used in the VQC. The layers will be applied in the given order.
            A double `StronglyEntanglingLayer` will be used if `None` is given.
        :param validation_set_size:
            A part of the training set that will be used for QMLModel validation.
            It should be from (0, 1). It will be used in each of the binary classifiers.
        :param rng_seed:
            A seed used for random weights initialization.
        """
        super().__init__(
            wires,
            device,
            optimizer,
            embedding_method,
            embedding_kwargs,
            layers,
            validation_set_size,
            rng_seed,
        )

        self.accuracy_threshold: float = accuracy_threshold
        self.n_classes = n_classes

        if not binary_classifiers:
            binary_classifiers = self._prepare_default_binary_classifiers(batch_size)

        self._binary_classifiers: Sequence[QNNBinaryClassifier] = binary_classifiers

    def _prepare_default_binary_classifiers(
        self, batch_size: int
    ) -> List[QNNBinaryClassifier]:
        """
        Prepares as set od default binary classifiers with the parameters specified
        during the class initialization.

        :param batch_size:
            Batch size to be used in the initialized binary classifiers.

        :return:
            List of the `QNNBinaryClassifier`s initialized with the given parameters.
        """
        binary_classifiers: List[QNNBinaryClassifier] = []

        for _ in range(self.n_classes):
            binary_classifiers.append(
                QNNBinaryClassifier(
                    wires=self.wires,
                    batch_size=batch_size,
                    embedding_method=self._embedding_method,
                    embedding_kwargs=self._embedding_kwargs,
                    layers=self._layers,
                    validation_set_size=self._validation_set_size,
                )
            )

        return binary_classifiers

    def seed(self, new_seed: int) -> None:
        """
        Sets up the new seed.

        :param new_seed:
            New seed to be applied to both the classifier and it's binary classifier
            parts.
        """
        self._rng_seed = new_seed

        for i in range(len(self._binary_classifiers)):
            self._binary_classifiers[i].seed(new_seed)

    def fit(
        self,
        X: Union[Sequence[Sequence[float]], NDArray[np.float64]],
        y: Optional[Sequence[ModelOutput]],
    ) -> "QNNClassifier":
        """
        The model training method. Essentially, it fits every binary classifier to the
        respective class.

        :param X:
            The lists of features of the objects that are used during the training.
        :param y:
            A list of outputs corresponding to the given lists of features.

        :raise AttributeError:
            If the device is not specified.
        :raise AttributeError:
            If the `y` is not specified.
        :raise AssertionError:
            If the number of provided binary classifiers and classes don't match.

        :return:
            Returns `self` after training.
        """

        if not self.dev:
            raise AttributeError("Specify the device (dev) before fitting.")

        if y is None:
            raise AttributeError("Missing y in supervised learning model.")

        # Check if there's a binary classifier for each class.
        unique_classes: np.ndarray = np.unique(y)

        if len(unique_classes) > len(self._binary_classifiers):
            raise AssertionError(
                "Numbers of provided binary classifiers and classes don't match!"
            )

        binary_classifier_accuracy_threshold: float = np.sqrt(self.accuracy_threshold)

        for classifier in self._binary_classifiers:
            classifier.accuracy_threshold = binary_classifier_accuracy_threshold

        # Fit each binary classifier to its respective class.
        for i in range(len(self._binary_classifiers)):
            classifier_classes = np.array(y)
            classifier_classes[classifier_classes != unique_classes[i]] = -1
            classifier_classes[classifier_classes == unique_classes[i]] = 1

            self._binary_classifiers[i].dev = self.dev
            self._binary_classifiers[i].fit(X, classifier_classes)

        return self

    def predict(self, features: Sequence[Sequence[float]]) -> Sequence[ModelOutput]:
        """
        Returns predictions of the model for the given features. In the case of
        `QuantumClassifier`, for given features, the predicted class corresponds to the
        index of the binary classifier which returns the highest expectation value of
        the `PauliZ` measurement on the first qubit.

        :param features:
            Features of the objects for which the model will predict the values.
        :return:
            Values predicted for given features.
        """
        predictions: List[int] = []

        for x in features:
            max_expectation: float = 0
            current_class: int = 0

            for i, classifier in enumerate(self._binary_classifiers):
                expectation: float = classifier.get_circuit_expectation_values([x])[0][
                    0
                ]

                if expectation > max_expectation:
                    max_expectation = expectation
                    current_class = i

            predictions.append(current_class)

        for classifier in self._binary_classifiers:
            classifier.circuit = None

        return predictions


class RBMClustering:
    """
    A class for performing clustering using Restricted Boltzmann Machine. The RBM can
    be trained using both classical and quantum algorithms. The quantum algorithm
    requires the D-Wave quantum annealer to be specified.
    """

    # TODO TR:  Remember to add the D-Wave handling.

    def __init__(
        self,
        lbae_input_shape: Tuple[int, ...],
        lbae_out_channels: int,
        lbae_n_layers: int,
        rbm_n_visible_neurons: int,
        rbm_n_hidden_neurons: int,
        n_gpus: int = 0,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        qubo_scale: float = 1.0,
        sampler: Optional[Sampler] = None,
        fireing_threshold: float = 0.8,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        self.lbae: LBAE = LBAE(
            input_size=lbae_input_shape[1:],  # TR: Notice shape reduction.
            out_channels=lbae_out_channels,
            latent_space_size=rbm_n_visible_neurons,
            num_layers=lbae_n_layers,
            quantize=True,  # Required, because it will be the input of the RBM.
        )

        self.rbm: RBM = RBM(
            num_visible=rbm_n_visible_neurons, num_hidden=rbm_n_hidden_neurons, rng=rng
        )

        self.n_gpus: int = n_gpus
        self.n_epochs: int = n_epochs
        self.learning_rate: float = learning_rate
        self.qubo_scale: float = qubo_scale

        self.sampler: Optional[Sampler] = sampler

        self.fireing_threshold: float = fireing_threshold

    def fit(
        self,
        data_loader: DataLoader[Tuple[Tensor, Tensor]],
    ) -> None:
        n_gpus: int = self.n_gpus if self.n_gpus > 0 else 1

        lbae_trainer: Trainer = Trainer(
            accelerator="cpu",  # TR TODO: Make it modifiable.
            num_nodes=n_gpus,
            max_epochs=self.n_epochs,
            deterministic=True,
        )

        print("LBAE training start.")
        lbae_trainer.fit(self.lbae, data_loader)
        print("LBAE training finished.")

        rbm_trainer: RBMTrainer
        # Pick a trainer depending on the existence of the sampler.
        if self.sampler is None:
            rbm_trainer = CD1Trainer(
                num_steps=self.n_epochs, learning_rate=self.learning_rate
            )
        else:
            rbm_trainer = AnnealingRBMTrainer(
                sampler=self.sampler,
                learning_rate=self.learning_rate,
                num_steps=self.n_epochs,
            )

        print("RBM training start.")
        rbm_trainer.fit(
            self.rbm, self._encoded_data_loader(data_loader, self.lbae.encoder)
        )
        print("RBM training finished.")

    @staticmethod
    def _encoded_data_loader(
        data_loader: DataLoader[Tuple[Tensor, Tensor]], encoder: LBAEEncoder
    ) -> Generator[Tuple[Tensor, Tensor], None, None]:
        """
        A generator that yields encoded data and targets from the given data loader. The
        data is encoded using the given encoder.

        :param data_loader:
            Data loader to be used for generating the encoded data.
        :param encoder:
            LBAE Encoder to be used for encoding the data.

        :return:
            Encoded data and targets.
        """
        while True:
            for _, (data, target) in enumerate(data_loader):
                yield encoder(data)[0], target

    def predict(self, x: Tensor) -> Tensor:
        """
        Returns the predicted class for the given input.

        :param x:
            Input to be classified.

        :return:
            Predicted class.
        """
        encoded_x: Tensor = self.lbae.encoder(x)[0]
        h_probs: NDArray[np.float32] = self.rbm.h_probs_given_v(
            encoded_x.detach().numpy()
        )[0]
        # print(h_probs)
        return Tensor([1 if p > self.fireing_threshold else 0 for p in h_probs])



"""
=============================================================================

    This module contains all the utilities used for data preprocessing (some initial
    analysis, normalization, ...).

=============================================================================

    Copyright 2023 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"


from copy import deepcopy
from typing import List, Sequence, Tuple, TypeVar, Union

from sklearn.preprocessing import MinMaxScaler

from aqmlator.data_acquisition import LearningDatum, SupervisedLearningDatum

AqmlatorDatum = TypeVar("AqmlatorDatum", LearningDatum, SupervisedLearningDatum)


def get_attributes(
    data: Sequence[LearningDatum],
) -> List[Tuple[Union[float, str], ...]]:
    """
    A method for extracting the attributes from a `Sequence` of the `LearningDatum`.

    :param data:
        A `Sequence` of `LearningDatum` from which the attributes will be extracted.

    :return:
        A list of attributes from given `LearningDatum` sequence. The order of the
        returned classes is the same as in the given data.
    """
    arguments: List[Tuple[Union[float, str], ...]] = []

    for datum in data:
        arguments.append(datum.datum_attributes)

    return arguments


def get_targets(
    data: Sequence[SupervisedLearningDatum],
) -> List[Union[float, str, int]]:
    """
    A method for extracting the classes (or target values) from a `Sequence` of the
    `SupervisedLearningDatum`.

    :param data:
        A `Sequence` of `SupervisedLearningDatum` from which the classes will be
        extracted.

    :return:
        A list of classes from given `SupervisedLearningDatum` sequence. The order of
        the returned classes is the same as in the given data.
    """
    targets: List[Union[float, str, int]] = []

    for datum in data:
        targets.append(datum.datum_target)

    return targets


class LearningDatumMinMaxScaler:
    """
    A class for rescaling the `datum_attributes` of the `LearningDatum`. It's meant
    to extend the required functionality of `sklearn.preprocessing.MinMaxScaler`.
    """

    def __init__(self, feature_range: Tuple[float, float] = (0, 1)):
        """
        A constructor for `LearningDatumMinMaxScaler`.

        :param feature_range:
            The range to which the values of learning datum attributes will be reduced.
            Notice that it's expected for the first value to be smaller than the
            second.
        """
        self._feature_range = feature_range

    def fit_transform(self, data: Sequence[AqmlatorDatum]) -> Sequence[AqmlatorDatum]:
        """
        Transforms the attributes of the `LearningDatum` objects so that they are
        within the bounds given by the `feature_range` specified in the constructor.

        :Note:
            We expect the datum_attributes to be of type `float` (not `str`, which is
            also formally allowed.

        :param data:
            A list of LearningDatum of which attributes will be transformed.

        :return:
            A list of transformed `LearningDatum`.
        """
        arguments: List[Tuple[Union[float, str], ...]] = get_attributes(data)

        fitted_attributes: List[Tuple[float, ...]] = MinMaxScaler(
            self._feature_range
        ).fit_transform(arguments)

        fitted_data: Sequence[AqmlatorDatum] = deepcopy(data)

        for i in range(len(data)):
            fitted_data[i].datum_attributes = fitted_attributes[i]

        return fitted_data



"""
=============================================================================

This file is a part of qbm4eo.rst project.

https://github.com/FeralQubits/qbm4eo

=============================================================================

It has been modified as a part of the EuroHPC PL project funded at the Smart Growth
Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

from math import prod
from typing import Any, Dict, List, Optional, Sequence

import torch
from torch import nn


class ResBlockDeConvPart(nn.Module):
    """
    A single part of the ResBlockDeConv class.
    """

    def __init__(
        self,
        channels: int,
        *args: Dict[str, Any],
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the ResBlockDeConvPart class.

        :param channels:
            The number of channels in the input and output.
        :param negative_slope:
            The negative slope of the LeakyReLU activation function.
        :param bias:
            Whether to use a bias term in the convolutional layer.
        :param args:
            Additional arguments to pass to the super class.
        :param kwargs:
            Additional keyword arguments to pass to the super class.
        """
        super().__init__(*args, **kwargs)

        self.subnet: nn.Sequential = nn.Sequential(
            nn.LeakyReLU(negative_slope),
            nn.ConvTranspose2d(
                channels, channels, kernel_size=3, stride=1, padding=1, bias=bias
            ),
            nn.BatchNorm2d(channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        The forward pass of the ResBlockDeConvPart class.

        :param x:
            The input tensor.

        :return:
            The output tensor.
        """
        return self.subnet(x)


class ResBlockDeConv(nn.Module):
    """
    A residual block for the decoder part of the LBAE model.
    """

    def __init__(
        self,
        channels: int,
        *args: Dict[str, Any],
        in_channels: Optional[int] = None,
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the ResBlockDeConv class.

        :param channels:
            The number of channels in the input and output.
        :param in_channels:
            The number of channels in the input. If None, defaults to the number of
            channels in the output. Defaults to None.
        :param negative_slope:
            The negative slope of the LeakyReLU activation function.
        :param bias:
            Whether to use a bias term in the convolutional layer.
        :param args:
            Additional arguments to pass to the super class.
        :param kwargs:
            Additional keyword arguments to pass to the super class.
        """
        super().__init__(*args, **kwargs)
        if in_channels is None:
            in_channels = channels

        self.initial_block: nn.Sequential = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels,
                channels,
                kernel_size=4,
                stride=2,
                padding=1,
                output_padding=0,
                bias=bias,
            ),
            nn.BatchNorm2d(channels),
        )

        self.middle_block: nn.Sequential = nn.Sequential(
            ResBlockDeConvPart(channels, negative_slope=negative_slope, bias=bias),
            ResBlockDeConvPart(channels, negative_slope=negative_slope, bias=bias),
        )

        self.negative_slope: float = negative_slope

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        The forward pass of the ResBlockDeConv class.

        :param x:
            The input tensor.

        :return:
            The output tensor.
        """
        x = self.initial_block(x)
        y: torch.Tensor = x
        x = self.middle_block(x)
        return nn.functional.leaky_relu(x + y, self.negative_slope)


class LBAEDecoder(nn.Module):
    """
    The decoder part of the LBAE model.
    """

    def __init__(
        self,
        input_size: Sequence[int],
        output_size: Sequence[int],
        latent_space_size: int,
        num_layers: int,
        *args: Dict[str, Any],
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the LBAEDecoder class.

        :param input_size:
            The size of the input tensor. This should be a tuple of the form
            (channels, height, width). The height and width should be divisible by
            2**num_layers.
        :param output_size:
            The size of the output tensor.
        :param latent_space_size:
            The size of the latent space. This is the size of the input to the linear
            layer.
        :param num_layers:
            The number of layers in the decoder. This is the number of ResBlockDeConv
            blocks.
        :param negative_slope:
            The negative slope of the LeakyReLU activation function.
        :param bias:
            Whether to use a bias term in the convolutional layer.
        :param args:
            Additional arguments to pass to the super class.
        :param kwargs:
            Additional keyword arguments to pass to the super class.
        """
        super().__init__(*args, **kwargs)

        self.input_size: Sequence[int] = input_size
        self.linear: nn.Linear = nn.Linear(latent_space_size, prod(input_size))

        layers: List[nn.Module] = []

        for i in range(num_layers):
            layers.append(
                ResBlockDeConv(
                    channels=input_size[0] // (2 ** (i + 1)),
                    in_channels=input_size[0] // (2**i),
                )
            )

        layers += [
            # Again, swapped input/output number of channels
            nn.ConvTranspose2d(
                input_size[0] // 2**num_layers,
                input_size[0] // 2**num_layers,
                kernel_size=4,
                stride=2,
                bias=bias,
                padding=1,
                output_padding=0,
            ),
            nn.BatchNorm2d(input_size[0] // 2**num_layers),
            nn.LeakyReLU(negative_slope),
            nn.ConvTranspose2d(
                input_size[0] // 2**num_layers,
                output_size[0],
                kernel_size=3,
                stride=1,
                padding=1,
                bias=bias,
            ),
        ]
        self.net: nn.Sequential = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        The forward pass of the LBAEDecoder class.

        :param x:
            The input tensor.

        :return:
            The output tensor of the decoder.
        """
        x = x.view(x.size(0), -1)
        x = self.linear(x)
        x = x.view(x.size(0), *self.input_size)

        x = self.net(x)
        x = torch.sigmoid(x)
        return x



"""
=============================================================================

This file is a part of qbm4eo.rst project.

https://github.com/FeralQubits/qbm4eo

=============================================================================

It has been modified as a part of the EuroHPC PL project funded at the Smart Growth
Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

from typing import Any, Dict, Tuple

import torch
from lightning.pytorch.core import LightningModule
from lightning.pytorch.utilities.types import OptimizerLRScheduler
from torch import Tensor, optim

from .decoder import LBAEDecoder
from .encoder import LBAEEncoder


def loss(xr: Tensor, x: Tensor) -> Tensor:
    """
        Loss function for LBAE. Uses MSE.

    :param xr:
        Reconstructed image.
    :param x:
        Original image.

    :return:
        MSE loss value.
    """
    return torch.nn.functional.mse_loss(xr, x, reduction="sum")


class LBAE(LightningModule):
    """
    A class implementing the Latent Bernoulli Autoencoder (LBAE) model.
    """

    def __init__(
        self,
        input_size: Tuple[int, ...],
        out_channels: int,
        latent_space_size: int,
        num_layers: int,
        quantize: bool,
        *args: Dict[str, Any],
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the LBAE class.

        :param input_size:
            The size of the input image.
        :param out_channels:
            The number of channels in the output image.
        :param latent_space_size:
            The size of the latent space.
        :param num_layers:
            The number of layers in the encoder and decoder.
        :param quantize:
            The epochs during which the output of the encoder should be quantized.
        :param args:
            Additional arguments.
        :param kwargs:
            Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)

        self.reference_image: Tensor = Tensor()

        self.save_hyperparameters(
            "input_size", "out_channels", "latent_space_size", "num_layers", "quantize"
        )

        self.encoder: LBAEEncoder = LBAEEncoder(
            input_size, out_channels, latent_space_size, num_layers, quantize
        )

        self.decoder: LBAEDecoder = LBAEDecoder(
            self.encoder.final_conv_size, input_size, latent_space_size, num_layers
        )

        self.epoch: int = 0

    def forward(self, x: Tensor) -> Tensor:
        """
        A forward pass through the LBAE model.

        :param x:
            The input image.

        :return:
            The reconstructed image.
        """

        z, _ = self.encoder(x)
        xr: Tensor = self.decoder(z)
        return xr

    def training_step(self, batch: Tuple[Tensor, Tensor], batch_idx: int) -> Tensor:
        """
        A training step for the LBAE model.

        :param batch:
            A batch of images.
        :param batch_idx:
            The index of the batch.

        :return:
            The loss value.
        """

        x: Tensor

        x = batch[0]

        if self.epoch == 0 and batch_idx == 0:
            self.reference_image = x[0:1, :, :, :]

        xr: Tensor = self.forward(x)
        l: Tensor = loss(xr.view(x.size()), x)

        self.log("loss", l, logger=True)

        return l

    def predict_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = ...  # type: ignore
    ) -> Any:
        """
        A function for predicting the output of the model.

        :param batch:
            A batch of images.
        :param batch_idx:
            The index of the batch.
        :param dataloader_idx:
            The index of the dataloader.

        :return:
            The output of the model, the input images, and the labels.
        """
        del batch_idx
        del dataloader_idx

        x: Tensor
        labels: Tensor

        x, labels = batch
        return self.forward(x), x, labels

    def test_step(self, batch: Any, batch_idx: int) -> Any:
        del batch_idx

        x, _ = batch
        xr = self.forward(x)
        return loss(xr.view(x.size()), x)

    def configure_optimizers(self) -> OptimizerLRScheduler:
        """
        A function for configuring the optimizers for the LBAE model.

        :return:
            A dictionary containing the configured optimizers.
        """
        return {"optimizer": optim.Adam(self.parameters(), lr=1e-3)}



"""
=============================================================================

    This module contains tests for the functionalities in the qml module.

=============================================================================

    Copyright 2022 ACK Cyfronet AGH. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

=============================================================================

    This work was supported by the EuroHPC PL project funded at the Smart Growth
    Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
    POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

__author__ = "Tomasz Rybotycki"


import abc
import os
import unittest
from typing import List, Optional, Sequence, Tuple, Type, Union

import dill
import lightning.pytorch.utilities.seed
import pennylane as qml
import torch
from dwave.samplers import RandomSampler
from numpy import isclose
from numpy.random import RandomState
from numpy.typing import NDArray
from pennylane import numpy as np
from pennylane.measurements import ExpectationMP
from pennylane.operation import Operation
from pennylane.templates import StronglyEntanglingLayers
from qiskit_ibm_provider import IBMProvider
from sklearn.datasets import (
    load_digits,
    make_classification,
    make_moons,
    make_regression,
)
from sklearn.metrics import rand_score
from torch import Tensor
from torch.utils.data import DataLoader

from aqmlator.qml import (
    QNNBinaryClassifier,
    QNNClassifier,
    QNNLinearRegression,
    QNNModel,
    QuantumKernelBinaryClassifier,
    RBMClustering,
)


class TestQNNModel(unittest.TestCase, abc.ABC):
    """
    A general `unittest.TestCase` class for QNN based QML models.
    """

    x: Sequence[Sequence[float]]
    y: Union[List[int], List[float]]

    model: QNNModel
    alternate_model: QNNModel

    def setUp(self) -> None:
        """
        Setup method for the `TestCase`. Should be overwritten by test classes.

        :note:
            TR: This is by default called before any test. One way to skip the test of this
            class if to set the skipping in the setUp. There may be a better way to do
            it though.
        """
        raise unittest.SkipTest("Skipping tests for abstract QNNModel class.")

    def tearDown(self) -> None:
        if os.path.isfile("qml_test.dil"):
            os.remove("qml_test.dil")

    @staticmethod
    def get_weights(model: torch.nn.Module) -> List[np.ndarray]:
        """
        Extract the weights from the given model.

        :param model:
            The model to extract the weights from.

        :return:
            The current weights of the model.
        """
        weights: List[np.ndarray] = []

        for _, param in model.named_parameters():
            weights.append(np.array(param.detach().numpy()))

        return weights

    def test_predict_run(self) -> None:
        """
        Tests if making predictions is possible.
        """
        self.model.predict(self.x)

    def test_fit_run(self) -> None:
        """
        Tests if the learning runs smoothly.
        """
        self.model.fit(self.x, self.y)

    def test_accuracy_increase(self) -> None:
        """
        Tests if the accuracy increases after short training.
        """
        initial_score: float = self.model.score(self.x, self.y)
        self.model.fit(self.x, self.y)
        final_score: float = self.model.score(self.x, self.y)
        self.assertTrue(
            final_score > initial_score,
            f"QNN Training: Initial score ({initial_score}) isn't worse than the final"
            f" score ({final_score})!",
        )

    def test_weights_change(self) -> None:
        """
        Tests if the weights change during the training.
        """
        initial_weights: Sequence[float] = self.model.weights
        self.model.fit(self.x, self.y)

        self.assertTrue(
            tuple(initial_weights) != tuple(self.model.weights),
            "Weights didn't change during the training!",
        )

    def test_results_dimensions(self) -> None:
        """
        Tests if the predictions have expected dimensions.
        """
        predictions: np.ndarray = self.model.predict(self.x)
        self.assertTrue(
            len(predictions) == len(self.x),
            f"Result dimensions are unexpected!({len(predictions)} != {len(self.x)}).",
        )

    def test_executions_number_growth(self) -> None:
        """
        Tests if the number of executions grows when the model is executed.
        """
        self.model.predict(self.x)
        self.assertTrue(
            self.model.n_executions() > 0, "The number of executions don't grow!"
        )

    def test_different_layers_predict_run(self) -> None:
        """
        Tests if making predictions is possible when different type of layers is used.
        """
        self.alternate_model.predict(self.x)

    def test_initial_serialization(self) -> None:
        """
        Tests if the model is serializable after the initialization.

        :note:
            This method can be (and is) used to check if the modified model is
            serializable.
        """

        model_score: float = self.model.score(self.x, self.y)

        with open("qml_test.dil", "wb") as f:
            dill.dump(self.model, f)

        with open("qml_test.dil", "rb") as f:
            loaded_model: QNNModel = dill.load(f)

        self.assertTrue(isclose(model_score, loaded_model.score(self.x, self.y)))

    def test_post_fit_serialization(self) -> None:
        """
        Tests if the model is serializable after fit.
        """
        self.model.fit(self.x, self.y)
        self.test_initial_serialization()

    def test_post_prediction_serialization(self) -> None:
        """
        Tests if the model is serializable after making prediction.
        """
        self.model.predict(self.x)
        self.test_initial_serialization()

    def test_torch_forward_run(self) -> None:
        """
        Tests if making predictions with torch classifier is possible.
        """
        model: torch.nn.Sequential = torch.nn.Sequential(self.model.get_torch_layer())
        model.forward(torch.tensor(self.x))

    def test_torch_results_dimension(self) -> None:
        """
        Tests if torch predictions have expected dimensions.
        """
        model: torch.nn.Sequential = torch.nn.Sequential(self.model.get_torch_layer())
        predictions: torch.Tensor = model.forward(torch.tensor(self.x))

        self.assertTrue(
            len(predictions) == len(self.x), "Torch predictions have unexpected shape."
        )

    def test_torch_different_layers_forward_run(self) -> None:
        """
        Tests if making predictions with torch is possible when different type of layers
        is used.
        """
        model: torch.nn.Sequential = torch.nn.Sequential(
            self.alternate_model.get_torch_layer()
        )
        model.forward(torch.tensor(self.x))


class TestQNNBinaryClassifier(TestQNNModel):
    """
    A `TestCase` class for the QNN-based binary classifier.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """
        # TR:   Changing the seed can cause problems with the `test_accuracy_increase`,
        #       as the number of training epochs is currently low for `seed = 1`.
        seed: int = 2
        noise: float = 0.1
        n_samples: int = 100
        accuracy_threshold: float = 0.85

        self.x, self.y = make_moons(
            n_samples=n_samples,
            shuffle=True,
            noise=noise,
            random_state=RandomState(seed),
        )

        for i in range(len(self.y)):
            if self.y[i] == 0:
                self.y[i] = -1

        n_qubits: int = 2

        dev: qml.Device = qml.device("lightning.qubit", wires=n_qubits)

        layers: List[Type[Operation]] = [
            StronglyEntanglingLayers
        ] * 3  # 3 StronglyEntanglingLayers

        alternate_layers: List[Type[Operation]] = [
            qml.templates.BasicEntanglerLayers
        ] * 2

        self.n_epochs: int = 2
        batch_size: int = 20

        self.model: QNNBinaryClassifier = QNNBinaryClassifier(
            wires=n_qubits,
            batch_size=batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=accuracy_threshold,
            layers=layers,
            device=dev,
        )

        self.alternate_model: QNNBinaryClassifier = QNNBinaryClassifier(
            wires=n_qubits,
            batch_size=batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=accuracy_threshold,
            layers=alternate_layers,
            device=dev,
        )


class TestQNNLinearRegressor(TestQNNModel):
    """
    A `TestCase` class for the QNN-based linear regressor.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """
        # TR:   Changing the seed can cause problems with the `test_accuracy_increase`,
        #       as the number of training epochs is currently low for `seed = 1`.
        seed: int = 2
        noise: float = 0.1
        n_samples: int = 100
        accuracy_threshold: float = 0.85

        self.x, self.y = make_regression(  # pylint: disable=unbalanced-tuple-unpacking
            n_samples=n_samples,
            n_features=2,
            shuffle=True,
            noise=noise,
            random_state=RandomState(seed),
        )

        n_qubits: int = 2
        dev: qml.Device = qml.device("lightning.qubit", wires=n_qubits)

        layers: List[Type[Operation]] = [
            StronglyEntanglingLayers
        ] * 3  # 3 StronglyEntanglingLayers

        alternate_layers: List[Type[Operation]] = [
            qml.templates.BasicEntanglerLayers
        ] * 2

        self.n_epochs: int = 3
        batch_size: int = 20

        self.model: QNNLinearRegression = QNNLinearRegression(
            wires=n_qubits,
            batch_size=batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=accuracy_threshold,
            layers=layers,
            device=dev,
        )

        self.alternate_model: QNNLinearRegression = QNNLinearRegression(
            wires=n_qubits,
            batch_size=batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=accuracy_threshold,
            layers=alternate_layers,
            device=dev,
        )


class TestQEKBinaryClassifier(unittest.TestCase):
    """
    A `TestCase` class for the QEK-based binary classifier.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """
        # TR:   Changing the seed can cause problems with the `test_accuracy_increase`,
        #       as the number of training epochs is currently minimal for `seed = 0`.
        seed: int = 0
        noise: float = 0.5
        n_samples: int = 15
        accuracy_threshold: float = 0.85

        self.x, self.y = make_moons(
            n_samples=n_samples,
            shuffle=True,
            noise=noise,
            random_state=RandomState(seed),
        )

        for i in range(len(self.y)):
            if self.y[i] == 0:
                self.y[i] = -1

        self.n_qubits: int = 2

        dev: qml.Device = qml.device("lightning.qubit", wires=self.n_qubits)

        layers: List[Type[Operation]] = [
            StronglyEntanglingLayers
        ] * 3  # 3 StronglyEntanglingLayers

        self.weights_length: int = 18

        alternate_layers: List[Type[Operation]] = [
            qml.templates.BasicEntanglerLayers
        ] * 3

        self.n_epochs: int = 1

        self.classifier: QuantumKernelBinaryClassifier = QuantumKernelBinaryClassifier(
            wires=self.n_qubits,
            n_epochs=self.n_epochs,
            accuracy_threshold=accuracy_threshold,
            layers=layers,
            device=dev,
        )

        self.alternate_classifier: QuantumKernelBinaryClassifier = (
            QuantumKernelBinaryClassifier(
                wires=self.n_qubits,
                n_epochs=self.n_epochs,
                accuracy_threshold=accuracy_threshold,
                layers=alternate_layers,
                device=dev,
            )
        )

    def tearDown(self) -> None:
        if os.path.isfile("qml_test.dil"):
            os.remove("qml_test.dil")

    def test_learning_and_predict_run(self) -> None:
        """
        Tests if fitting and making predictions is possible.
        """
        self.classifier.fit(self.x, self.y)
        self.classifier.predict(self.x)

    def test_accuracy_increase(self) -> None:
        """
        Tests if the accuracy increases after short training.
        """
        self.classifier.fit(self.x, self.y)
        initial_accuracy: float = self.classifier.score(self.x, self.y)

        self.classifier.n_epochs = 2  # Minimal required number in this setup.
        self.classifier.fit(self.x, self.y)
        accuracy: float = self.classifier.score(self.x, self.y)

        self.assertTrue(
            initial_accuracy < accuracy,
            f"Initial accuracy ({initial_accuracy}) didn't increase ({accuracy}) after "
            f"training.",
        )

    def test_weights_change(self) -> None:
        """
        Tests if the weights change during the training.
        """
        initial_weights: Sequence[float] = self.classifier.weights
        self.classifier.fit(self.x, self.y)

        self.assertTrue(
            tuple(initial_weights) != tuple(self.classifier.weights),
            "Weights didn't change during the training!",
        )

    def test_results_dimension(self) -> None:
        """
        Tests if the predictions have expected dimensions.
        """
        self.classifier.fit(self.x, self.y)
        predictions: np.ndarray = np.array(self.classifier.predict(self.x))
        self.assertTrue(
            predictions.shape == (len(self.x),),
            "QuantumKernelBinaryClassifier predictions have unexpected shape.",
        )

    def test_executions_number_growth(self) -> None:
        """
        Tests if the number of executions grows when the model is executed.
        """
        self.classifier.fit(self.x, self.y)
        self.classifier.predict(self.x)
        self.assertTrue(
            self.classifier.n_executions() > 0, "The number of executions don't grow!"
        )

    def test_different_layers_learning_and_predict_run(
        self,
    ) -> None:
        """
        Tests if making predictions is possible when different type of layers is used.
        """
        self.alternate_classifier.fit(self.x, self.y)
        self.alternate_classifier.predict(self.x)

    def test_transform_run(self) -> None:
        """
        Checks if `classifier.transform` runs.
        """
        self.classifier.fit(self.x, self.y)
        self.classifier.transform(self.x)

    def test_transform_dimension(self) -> None:
        """
        Checks if `classifier.transform` dimensions are as expected.
        """
        self.classifier.fit(self.x, self.y)
        mapped_x: List[ExpectationMP] = self.classifier.transform(self.x)

        self.assertTrue(
            len(mapped_x) == len(self.x),
            f"The results_reconstruction number is incorrect ({len(mapped_x)} != {len(self.x)})!",
        )

        for x in mapped_x:
            self.assertTrue(
                len(np.array(x)) == self.n_qubits,
                f"Dimension of the results_reconstruction is incorrect! ({len(np.array(x))} !="
                f" {self.n_qubits})",
            )

    def _test_serialization(self) -> None:
        """
        Tests if the model is serializable after the initialization.

        :note:
            This method can be (and is) used to check if the modified model is
            serializable.
        """

        model_score: float = self.classifier.score(self.x, self.y)

        with open("qml_test.dil", "wb") as f:
            dill.dump(self.classifier, f)

        with open("qml_test.dil", "rb") as f:
            loaded_model: QNNModel = dill.load(f)

        self.assertTrue(isclose(model_score, loaded_model.score(self.x, self.y)))

    def test_post_fit_serialization(self) -> None:
        """
        Tests if the model is serializable after fit.
        """
        self.classifier.fit(self.x, self.y)
        self._test_serialization()

    def test_post_fit_prediction_serialization(self) -> None:
        """
        Tests if the model is serializable after making prediction.
        """
        self.classifier.fit(self.x, self.y)
        self.classifier.predict(self.x)
        self._test_serialization()


class TestQuantumClassifier(unittest.TestCase):
    """
    A `TestCase` class for general quantum classifier.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """

        self.n_samples: int = 50
        seed: int = 0
        n_classes: int = 3
        n_epochs: int = 2
        batch_size: int = 10
        n_features: int = 2

        self.X: Sequence[Sequence[float]]
        self.y: Sequence[int]

        self.X, self.y = make_classification(
            n_samples=self.n_samples,
            n_features=n_features,
            n_classes=n_classes,
            n_redundant=0,
            n_clusters_per_class=1,
            random_state=RandomState(seed),
        )

        dev: qml.Device = qml.device("lightning.qubit", wires=n_features)

        classifiers: List[QNNBinaryClassifier] = [
            QNNBinaryClassifier(
                wires=n_features, batch_size=batch_size, n_epochs=n_epochs, device=dev
            )
            for _ in range(n_classes)
        ]

        self.classifier: QNNClassifier = QNNClassifier(
            wires=n_features,
            binary_classifiers=classifiers,
            n_classes=n_classes,
            device=dev,
        )

    def tearDown(self) -> None:
        if os.path.isfile("qml_test.dil"):
            os.remove("qml_test.dil")

    def test_predict_run(self) -> None:
        """
        Tests if making predictions is possible.
        """
        self.classifier.predict(self.X)

    def test_fit_run(self) -> None:
        """
        Tests if classifier fitting is possible.
        """
        self.classifier.fit(self.X, self.y)

    def test_results_dimensions(self) -> None:
        """
        Tests if the dimension of the results_reconstruction returned by the classifier is correct.
        """
        results: Sequence[int] = self.classifier.predict(self.X)
        self.assertTrue(len(results) == self.n_samples)

    def test_accuracy_increase(self) -> None:
        """
        Tests if the classifier accuracy increase after the training.
        """
        initial_accuracy: float = np.mean(
            [int(i == j) for i, j in zip(self.y, self.classifier.predict(self.X))]
        )
        self.classifier.fit(self.X, self.y)
        final_accuracy: float = np.mean(
            [int(i == j) for i, j in zip(self.classifier.predict(self.X), self.y)]
        )
        self.assertTrue(initial_accuracy < final_accuracy)

    def test_initial_serialization(self) -> None:
        """
        Tests if the model is serializable after the initialization.

        :note:
            This method can be (and is) used to check if the modified model is
            serializable.
        """

        model_score: float = self.classifier.score(self.X, self.y)

        with open("qml_test.dil", "wb") as f:
            dill.dump(self.classifier, f)

        with open("qml_test.dil", "rb") as f:
            loaded_model: QNNModel = dill.load(f)

        self.assertTrue(isclose(model_score, loaded_model.score(self.X, self.y)))

    def test_post_fit_serialization(self) -> None:
        """
        Tests if the model is serializable after fit.
        """
        self.classifier.fit(self.X, self.y)
        self.test_initial_serialization()

    def test_post_prediction_serialization(self) -> None:
        """
        Tests if the model is serializable after making prediction.
        """
        self.classifier.predict(self.X)
        self.test_initial_serialization()


# TODO TR: Think of a less general case for this class.
class TestIBMQDevicesHandling(unittest.TestCase):
    """
    A class for testing if the qml models work as intended on IBM devices.
    """

    def setUp(self) -> None:
        """
        Sets up the tests. Called before every test.
        """
        n_samples: int = 50
        seed: int = 0

        self.n_features: int = 3
        self.n_classes: int = 2
        self.noise: float = 0.1
        self.batch_size: int = n_samples // 5
        self.n_epochs: int = 1
        self.accuracy_threshold: float = 0.8

        self.class_X: Sequence[Sequence[float]]
        self.class_y: Sequence[int]

        self.class_X, self.class_y = make_classification(
            n_samples=n_samples,
            n_features=self.n_features,
            n_classes=self.n_classes,
            n_redundant=0,
            n_clusters_per_class=1,
            random_state=RandomState(seed),
        )

        self.regression_X: Sequence[Sequence[float]]
        self.regression_y: Sequence[float]

        (  # pylint: disable=unbalanced-tuple-unpacking
            self.regression_X,
            self.regression_y,
        ) = make_regression(
            n_samples=n_samples,
            n_features=self.n_features,
            shuffle=True,
            noise=self.noise,
            random_state=RandomState(seed),
        )

        provider = IBMProvider(instance="ibm-q/open/main")
        backends = provider.backends()

        for i in range(len(backends)):
            if (
                "simulator" in str(backends[i]).lower()
                or backends[i].configuration().n_qubits < 3
            ):
                continue
            backend = backends[i]
            self.n_qubits: int = backend.configuration().n_qubits
            break

        config = backend.configuration()

        self.coupling_map: List[Sequence[int]] = config.coupling_map

        self.dev: qml.Device = qml.device(
            "qiskit.aer",
            wires=self.n_features,
            backend="aer_simulator_statevector",
        )

        self.coupled_dev = qml.device(
            "qiskit.aer",
            wires=self.n_features,
            backend="aer_simulator_statevector",
            coupling_map=self.coupling_map,
        )

        self.layers: List[Type[Operation]] = [
            StronglyEntanglingLayers
        ] * 3  # 3 StronglyEntanglingLayers

    def _proceed_with_qek_classifier_test(
        self,
        coupling_map: Optional[List[Sequence[int]]] = None,
        dev: Optional[qml.Device] = None,
    ) -> None:
        """
        A common part of all the QEK Classifier-related tests. Test is passed if the
        fitting don't crash.

        :param coupling_map:
            A coupling map to be applied when applying the VQC.
        :param dev:
            A device to run the VQC on.
        """
        if not dev:
            dev = self.dev

        qek_classifier: QuantumKernelBinaryClassifier = QuantumKernelBinaryClassifier(
            wires=self.n_features,
            n_epochs=self.n_epochs,
            accuracy_threshold=self.accuracy_threshold,
            layers=self.layers,
            device=dev,
            coupling_map=coupling_map,
        )
        qek_classifier.fit(self.class_X, self.class_y)

    def _proceed_with_qnn_regressor_test(
        self,
        coupling_map: Optional[List[Sequence[int]]] = None,
        dev: Optional[qml.Device] = None,
    ) -> None:
        """
        A common part of all the QNN Regressor-related tests. Test is passed if the
        fitting don't crash.

        :param coupling_map:
            A coupling map to be applied when applying the VQC.
        :param dev:
            A device to run the VQC on.
        """
        if not dev:
            dev = self.dev

        qnn_regressor: QNNLinearRegression = QNNLinearRegression(
            wires=self.n_features,
            batch_size=self.batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=self.accuracy_threshold,
            layers=self.layers,
            device=dev,
            coupling_map=coupling_map,
        )
        qnn_regressor.fit(self.regression_X, self.regression_y)

    def _proceed_wth_qnn_classifier_test(
        self,
        coupling_map: Optional[List[Sequence[int]]] = None,
        dev: Optional[qml.Device] = None,
    ) -> None:
        """
        A common part of all the QNN Classifier-related tests. Test is passed if the
        fitting don't crash.

        :param coupling_map:
            A coupling map to be applied when applying the VQC.
        :param dev:
            A device to run the VQC on.
        """
        if not dev:
            dev = self.dev

        qnn_classifier: QNNBinaryClassifier = QNNBinaryClassifier(
            wires=self.n_features,
            batch_size=self.batch_size,
            n_epochs=self.n_epochs,
            accuracy_threshold=self.accuracy_threshold,
            layers=self.layers,
            device=dev,
            coupling_map=coupling_map,
        )

        qnn_classifier.fit(self.class_X, self.class_y)

    def test_qek_classifier_on_qiskit_simulator(self) -> None:
        """
        Tests if the QEK classifier works correctly on the unconstrained IBMQ device
        simulator.
        """
        self._proceed_with_qek_classifier_test()

    def test_qnn_classifier_on_qiskit_simulator(self) -> None:
        """
        Tests if the QNN classifier works correctly on the unconstrained IBMQ device
        simulator.
        """
        self._proceed_wth_qnn_classifier_test()

    def test_qnn_regressor_on_qiskit_simulator(self) -> None:
        """
        Tests if the QNN regressor works correctly on the unconstrained IBMQ device
        simulator.
        """
        self._proceed_with_qnn_regressor_test()

    def test_qek_classifier_with_coupling(self) -> None:
        """
        Tests if the QEK classifier works correctly with the coupling map applied
        on the unconstrained IBMQ device simulator.
        """
        self._proceed_with_qek_classifier_test(self.coupling_map)

    def test_qnn_classifier_with_coupling(self) -> None:
        """
        Tests if the QNN classifier works correctly with the coupling map applied
        on the unconstrained IBMQ device simulator.
        """
        self._proceed_wth_qnn_classifier_test(self.coupling_map)

    def test_qnn_regressor_with_coupling(self) -> None:
        """
        Tests if the QNN regressor works correctly with the coupling map applied
        on the unconstrained IBMQ device simulator.
        """
        self._proceed_with_qnn_regressor_test(self.coupling_map)

    def test_qnn_classifier_on_coupled_device(self) -> None:
        """
        Tests if the QNN classifier works correctly with the coupling map applied
        on the real IBMQ device simulator.
        """
        self._proceed_wth_qnn_classifier_test(
            dev=self.coupled_dev, coupling_map=self.coupling_map
        )

    def test_qnn_regressor_on_coupled_device(self) -> None:
        """
        Tests if the QNN regressor works correctly with the coupling map applied
        on the real IBMQ device simulator.
        """
        self._proceed_with_qnn_regressor_test(
            dev=self.coupled_dev, coupling_map=self.coupling_map
        )

    def test_qek_classifier_on_coupled_device(self) -> None:
        """
        Tests if the QNN classifier works correctly with the coupling map applied
        on the real IBMQ device simulator.
        """
        self._proceed_with_qek_classifier_test(
            dev=self.coupled_dev, coupling_map=self.coupling_map
        )


class TestRBMClustering(unittest.TestCase):
    """
    Tests for the RBMClustering class.
    """

    def setUp(self) -> None:
        """
        Sets up the test case.
        """
        lightning.pytorch.seed_everything(0, workers=True)  # Fix the seed.

        lbae_input_size: Tuple[int, ...] = (1, 1, 8, 8)
        lbae_out_channels: int = 8
        n_layers: int = 2
        rbm_n_visible_neurons: int = 16
        rbm_n_hidden_neurons: int = 10

        n_epochs: int = 3
        batch_size: int = 10
        n_classes: int = 10

        X: NDArray[np.int]
        self.y: NDArray[np.int]

        X, self.y = load_digits(n_class=n_classes, return_X_y=True)
        X = X.reshape((1797, 1, 8, 8))

        self.X_tensor: Tensor = torch.Tensor(X)
        self.X_tensor /= 16  # Inputs have values from 0 to 16. Rescale to 0-1.

        # Use list instead of CustomDataset, as it implements everything that Dataset
        # is supposed to implement.
        dataset: List[Tuple[Tensor, Tensor]] = []

        for i in range(len(X)):
            dataset.append((self.X_tensor[i], self.y[i]))

        # Ignore mypy problem with the type of the dataset.
        self.data_loader: DataLoader[Tuple[Tensor, Tensor]] = DataLoader(
            dataset,  # type: ignore
            batch_size=batch_size,
            shuffle=True,
        )

        self.rbm_clustering: RBMClustering = RBMClustering(
            lbae_input_shape=lbae_input_size,
            lbae_out_channels=lbae_out_channels,
            lbae_n_layers=n_layers,
            rbm_n_visible_neurons=rbm_n_visible_neurons,
            rbm_n_hidden_neurons=rbm_n_hidden_neurons,
            n_epochs=n_epochs,
            rng=np.random.default_rng(seed=42),
            fireing_threshold=0.5,
        )

        self.x = self.X_tensor[0].view(1, 1, 8, 8)

    def tearDown(self) -> None:
        """
        Tears down the test case.
        """
        pass

    def _test_fit_run(self) -> None:
        """
        A common part of the tests for the fit method.
        """
        self.rbm_clustering.fit(self.data_loader)
        # self.assertTrue(True)

    def test_classical_clustering_fit_run(self) -> None:
        """
        Tests if the classically trained (using CD1 algorithm) RBMs fits without error.
        CD1 algorithm is used by default, when no sampler is specified.
        """
        self._test_fit_run()

    def test_sampler_clustering_fit_run(self) -> None:
        """
        Tests if the RBMs fits without error, when the sampler is specified.
        """
        sampler: RandomSampler = RandomSampler()
        self.rbm_clustering.sampler = sampler
        self._test_fit_run()

    def test_rbm_predict(self) -> None:
        """
        Tests if the RBMClustering predict method runs and returns binary values.
        """
        prediction: Tensor = self.rbm_clustering.predict(self.x)

        for val in prediction:
            self.assertTrue(val in (0, 1))

    def test_clustering_accuracy_increase(self) -> None:
        """
        Tests if the accuracy of the clustering increases after the (classical)
        training (which it should).
        """
        predictions: List[int] = []

        def simple_hash(t: Tensor) -> int:
            return sum(p * 2**i for i, p in enumerate(t))

        for x in self.X_tensor:
            predictions.append(
                simple_hash(self.rbm_clustering.predict(x.view(1, 1, 8, 8)))
            )

        initial_score: float = rand_score(self.y, predictions)

        predictions.clear()

        self.rbm_clustering.fit(self.data_loader)

        for x in self.X_tensor:
            predictions.append(
                simple_hash(self.rbm_clustering.predict(x.view(1, 1, 8, 8)))
            )

        final_score: float = rand_score(self.y, predictions)

        self.assertGreater(final_score, initial_score)



"""
Tests for the rbm.py module.
"""
import io

import dimod
import numpy as np
import torch

from qbm4eo.rbm import RBM, qubo_from_rbm_coefficients


class TestRBMToQUBOConversion:
    """
    Tests for the conversion of an RBM to a QUBO.
    """

    def test_qubo_is_correctly_constructed_for_single_unit_cell_rbm(self) -> None:
        weights = torch.tensor(
            [
                [0.5, 0.4, 0.2, 0.3],
                [0.4, -0.2, 0.01, 1.1],
                [-0.25, 0.5, 0.6, -0.7],
                [1.0, 0.5, -0.25, -0.3],
            ],
            dtype=torch.float64,
        )
        v_bias = torch.tensor([0.1, -0.2, 0.3, 0.4], dtype=torch.float64)
        h_bias = torch.tensor([-0.25, -0.3, 1.0, 0.25], dtype=torch.float64)
        expected_bqm = dimod.BQM(
            {0: 0.1, 1: -0.2, 2: 0.3, 3: 0.4, 4: -0.25, 5: -0.3, 6: 1.0, 7: 0.25},
            {
                (0, 4): 0.5,
                (0, 5): 0.4,
                (0, 6): 0.2,
                (0, 7): 0.3,
                (1, 4): 0.4,
                (1, 5): -0.2,
                (1, 6): 0.01,
                (1, 7): 1.1,
                (2, 4): -0.25,
                (2, 5): 0.5,
                (2, 6): 0.6,
                (2, 7): -0.7,
                (3, 4): 1.0,
                (3, 5): 0.5,
                (3, 6): -0.25,
                (3, 7): -0.3,
            },
            offset=0.0,
            vartype="BINARY",
        )

        bqm = qubo_from_rbm_coefficients(weights, v_bias, h_bias)

        assert bqm == expected_bqm


class TestRBM:
    """
    Tests for the RBM class.
    """

    def test_loading_saved_rbm_gives_back_original_rbm(self) -> None:
        original_rbm: RBM = RBM(15, 13)
        dummy_file: io.BytesIO = io.BytesIO()
        original_rbm.save(dummy_file)
        dummy_file.seek(0)

        loaded_rbm: RBM = RBM.load(dummy_file)

        # Perform the assertions.
        np.testing.assert_array_equal(original_rbm.weights, loaded_rbm.weights)
        np.testing.assert_array_equal(original_rbm.v_bias, loaded_rbm.v_bias)
        np.testing.assert_array_equal(original_rbm.h_bias, loaded_rbm.h_bias)

        assert original_rbm.num_visible == loaded_rbm.num_visible
        assert original_rbm.num_hidden == loaded_rbm.num_hidden



from setuptools import find_packages, setup

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="aqmlator",
    version="0.1.1",
    author="Tomasz Rybotycki",
    author_email="rybotycki.tomasz+aqmlator@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="A package for auto quantum machine learning-izing your experiments!",
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    license="Apache License 2.0.",
)
